"""
old_analytics for fitting vol for log sv model
"""
import numpy as np
import pandas as pd
import qis
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Tuple, Optional, List
from enum import Enum

import option_chain_analytics.fitters.vol_beta_fitter as fvb
import option_chain_analytics.pricers.bsm as bsm
from option_chain_analytics.option_chain import SliceColumn, UnderlyingColumn, ExpirySlice, SlicesChain


def run_chain_vol_fit(chain: SlicesChain,
                      delta_bounds: Tuple[Optional[float], Optional[float]] = (-0.05, 0.05)
                      ) -> pd.DataFrame:
    fit_params = {}
    for slice_id, slice_data in chain.expiry_slices.items():
        vols, strikes = slice_data.get_bid_mark_ask_vols(delta_bounds=delta_bounds, is_filtered=False)
        print(f"slice_id={slice_id}, forward={slice_data.forward}")
        print(vols)

        if vols is not None:
            mid_vols = vols[SliceColumn.MARK_IV].to_numpy()
            log_strikes = np.log(strikes.to_numpy() / slice_data.forward)

            fit_params[slice_id] = fvb.fit_logsv_ivols(log_strikes=log_strikes,
                                                       mid_vols=mid_vols,
                                                       ttm=slice_data.get_ttm(),
                                                       is_vega_weights=True)
    fit_params = pd.DataFrame.from_dict(fit_params, orient='index')
    return fit_params


def run_chain_report(chain: SlicesChain,
                     delta_bounds: Tuple[Optional[float], Optional[float]] = (-0.25, 0.25)
                     ) -> List[plt.Figure]:

    fit_params = {}
    figs = []
    for slice_id, slice_data in chain.expiry_slices.items():
        vols, strikes = slice_data.get_bid_mark_ask_vols(delta_bounds=delta_bounds, is_filtered=False)
        print(vols)
        if vols is not None:
            vols = vols.dropna(how='any', axis=0)
            strikes = vols.index
            print(f"slice_id={slice_id}")
            print(vols)

            mid_vols = vols[SliceColumn.MARK_IV].to_numpy()
            log_strikes = np.log(strikes.to_numpy() / slice_data.forward)

            fit_params[slice_id] = fvb.fit_logsv_ivols(log_strikes=log_strikes,
                                                       mid_vols=mid_vols,
                                                       ttm=slice_data.get_ttm(),
                                                       is_vega_weights=True)
            model_vols = fvb.calc_logsv_ivols(log_strikes=log_strikes, **fit_params[slice_id])

            df = pd.concat([vols, pd.Series(model_vols, index=vols.index, name='Fit')], axis=1)

            with sns.axes_style("darkgrid"):
                fig, ax = plt.subplots(1, 1, figsize=(10, 7))
                qis.plot_line(df=df,
                              title=f"{slice_id}",
                              yvar_format='{:,.4%}',
                              ax=ax)
            figs.append(fig)

    fit_params = pd.DataFrame.from_dict(fit_params, orient='index')
    print(fit_params)

    return figs


class UnitTests(Enum):
    DERIBIT_FIT = 1
    YAHOO_FIT = 2


def run_unit_test(unit_test: UnitTests):

    # to get options data
    from option_chain_analytics.data.chain_loader_from_ts import create_chain_from_from_options_dfs
    from option_chain_analytics.data.chain_ts import OptionsDataDFs
    import local_path as local_path

    if unit_test == UnitTests.DERIBIT_FIT:
        from option_chain_analytics.public_apis.deribit import load_deribit_contract_ts_data
        options_data_dfs = OptionsDataDFs(**load_deribit_contract_ts_data(ticker='ETH'))
        time_index = options_data_dfs.get_timeindex()
        print(f"time_index={time_index}")
        value_time = pd.Timestamp('2023-11-03 17:30:34.270745+00:00')
        chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)
        fit_params = run_chain_vol_fit(chain=chain)
        print(fit_params)

    elif unit_test == UnitTests.YAHOO_FIT:
        from option_chain_analytics.public_apis.yahoo import load_contract_ts_data

        ticker = 'USO'
        value_time = pd.Timestamp('2023-11-24 17:21:17.248659+00:00')

        ticker = 'SPY'
        value_time = pd.Timestamp('2023-11-26 20:00:00+00:00')

        options_data_dfs = OptionsDataDFs(**load_contract_ts_data(ticker=ticker))
        time_index = options_data_dfs.get_timeindex()
        print(f"time_index={time_index}")
        chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)
        figs = run_chain_report(chain=chain)
        qis.save_figs_to_pdf(figs, file_name=f"{ticker}_vol_fit", local_path=local_path.get_output_path())

        # run_chain_report(chain=chain)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.YAHOO_FIT

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




"""
dependence on stochvolmodels for model calibration
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import qis
from typing import Dict, Tuple, Optional
from numba.typed import List
from enum import Enum

# analytics
from stochvolmodels.data.option_chain import OptionChain

# chain
from option_chain_analytics import OptionsDataDFs, create_chain_from_from_options_dfs
from option_chain_analytics.option_chain import SliceColumn, SlicesChain


pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)


def generate_vol_chain_np(chain: SlicesChain,
                          value_time: pd.Timestamp,
                          days_map: Dict[str, int] = {'1w': 7, '1m': 21},
                          delta_bounds: Tuple[Optional[float], Optional[float]] = (-0.1, 0.1),
                          is_filtered: bool = True
                          ) -> OptionChain:
    # generate table snapshot of vol data: index= days, columns = deltas

    ttms, future_prices, discfactors = List(), List(), List()
    optiontypes_ttms, strikes_ttms = List(), List()
    bid_ivs, ask_ivs = List(), List()
    bid_prices, ask_prices = List(), List()
    slice_ids = []
    for label, day in days_map.items():
        next_date = value_time + pd.DateOffset(days=day)  # if overlapping next date will be last avilable maturity
        slice_date = chain.get_next_slice_after_date(mat_date=next_date)
        slice_t = chain.expiry_slices[slice_date]
        df = slice_t.get_joint_slice(delta_bounds=delta_bounds, is_filtered=is_filtered)

        slice_ids.append(f"{label}: {slice_t.expiry_id}")
        ttms.append(slice_t.get_ttm())
        future_prices.append(slice_t.get_future_price())
        discfactors.append(1.0)
        strikes_ttms.append(df.index.to_numpy())
        optiontypes_ttms.append(df[SliceColumn.OPTION_TYPE].to_numpy(dtype=str))
        bid_ivs.append(df[SliceColumn.BID_IV].to_numpy())
        ask_ivs.append(df[SliceColumn.ASK_IV].to_numpy())
        bid_prices.append(df[SliceColumn.BID_PRICE].to_numpy())
        ask_prices.append(df[SliceColumn.ASK_PRICE].to_numpy())

    out = OptionChain(ttms=np.array(ttms),
                      forwards=np.array(future_prices),
                      discfactors=np.array(discfactors),
                      ids=np.array(slice_ids),
                      strikes_ttms=strikes_ttms,
                      optiontypes_ttms=optiontypes_ttms,
                      bid_ivs=bid_ivs,
                      ask_ivs=ask_ivs,
                      bid_prices=bid_prices,
                      ask_prices=ask_prices)
    return out


def load_option_chain(options_data_dfs: OptionsDataDFs,
                      value_time: pd.Timestamp = pd.Timestamp('2023-02-06 08:00:00+00:00'),
                      days_map: Dict[str, int] = {'1w': 7, '1m': 21},
                      delta_bounds: Tuple[Optional[float], Optional[float]] = (-0.1, 0.1),
                      is_filtered: bool = True
                      ) -> Optional[OptionChain]:
    chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)
    if chain is not None:
        option_chain = generate_vol_chain_np(chain=chain,
                                             value_time=value_time,
                                             days_map=days_map,
                                             delta_bounds=delta_bounds,
                                             is_filtered=is_filtered)
    else:
        option_chain = None

    return option_chain


def sample_option_chain_at_times(options_data_dfs: OptionsDataDFs,
                                 time_period: qis.TimePeriod,
                                 freq: str = 'W-FRI',
                                 days_map: Dict[str, int] = {'1w': 7, '1m': 21},
                                 hour_offset: int = 8,
                                 delta_bounds: Tuple[Optional[float], Optional[float]] = (-0.1, 0.1),
                                 ) -> Dict[pd.Timestamp, OptionChain]:
    value_times = qis.generate_dates_schedule(time_period=time_period,
                                              freq=freq,
                                              hour_offset=hour_offset)
    option_chains = {}
    for value_time in value_times:
        option_chains[value_time] = load_option_chain(options_data_dfs=options_data_dfs,
                                                      value_time=value_time,
                                                      days_map=days_map,
                                                      delta_bounds=delta_bounds,
                                                      is_filtered=True)
    return option_chains


class UnitTests(Enum):
    PRINT_CHAIN_DATA = 1
    GENERATE_VOL_CHAIN_NP = 2
    SAMPLE_CHAIN_AT_TIMES = 3


def run_unit_test(unit_test: UnitTests):

    import local_path as local_path

    ticker = 'BTC'  # BTC, ETH
    value_time = pd.Timestamp('2021-10-21 08:00:00+00:00')
    value_time = pd.Timestamp('2023-10-06 08:00:00+00:00')

    from option_chain_analytics.ts_loaders import ts_data_loader_wrapper
    options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker))
    options_data_dfs.get_start_end_date().print()
    chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)

    if unit_test == UnitTests.PRINT_CHAIN_DATA:
        for expiry, eslice in chain.expiry_slices.items():
            eslice.print()

    elif unit_test == UnitTests.GENERATE_VOL_CHAIN_NP:
        option_chain = generate_vol_chain_np(chain=chain,
                                             value_time=value_time,
                                             days_map={'1w': 7},
                                             delta_bounds=(-0.1, 0.1),
                                             is_filtered=True)
        option_chain.print()
        skews = option_chain.get_chain_skews(delta=0.35)
        print(skews)

    elif unit_test == UnitTests.SAMPLE_CHAIN_AT_TIMES:
        time_period = qis.TimePeriod('01Jan2023', '31Jan2023', tz='UTC')
        option_chains = sample_option_chain_at_times(options_data_dfs=options_data_dfs,
                                                     time_period=time_period,
                                                     freq='W-FRI',
                                                     hour_offset=9
                                                     )
        for key, chain in option_chains.items():
            print(f"{key}")
            print(chain)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.SAMPLE_CHAIN_AT_TIMES

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
collector of aligned time series data for option
options data is split in frames
"""
from __future__ import annotations

# package
import pandas as pd
import numpy as np
from dataclasses import dataclass, field
from typing import Literal, List, Union, Optional
from enum import Enum
import qis
from qis import TimePeriod

# analytics
from option_chain_analytics.option_chain import SliceColumn
from option_chain_analytics.data.config import get_ttm_from_future_ticker, FutureTickerConfig


@dataclass
class ChainTs:
    """
    df collection for options and futures time series
    chain_ts must contain SliceColumn.EXCHANGE_TIME.value
    the most frequent request will be get df[contracts] at given timestamp
    to expedite request we add groupby[SliceColumn.EXCHANGE_TIME.value]
    """
    chain_ts: pd.DataFrame
    spot_data: pd.DataFrame
    ticker: str

    def __post_init__(self):
        self.chain_ts_group_by_time = self.chain_ts.groupby(SliceColumn.EXCHANGE_TIME.value)
        self.ewm_vol: Optional[pd.Series] = None

    def get_contract_data(self, contact: str) -> pd.DataFrame:
        df = self.chain_ts.loc[self.chain_ts[SliceColumn.CONTRACT.value] == contact, :]
        return df

    def print(self):
        print('chain_ts')
        print(self.chain_ts)
        print('spot_data')
        print(self.spot_data)

    def get_timeindex(self) -> pd.DatetimeIndex:
        return pd.DatetimeIndex(self.chain_ts_group_by_time.groups.keys())

    def get_start_end_date(self) -> TimePeriod:
        times = self.get_timeindex()
        return qis.TimePeriod(times.min(), times.max())

    def get_spot_data(self, time_period: TimePeriod = None) -> pd.DataFrame:
        df = self.spot_data
        if time_period is not None:
            df = time_period.locate(df)
        return df

    def get_time_slice(self,
                       timestamp: pd.Timestamp,
                       contracts: List[str] = None
                       ) -> pd.DataFrame:
        """
         get non-nan time slice from series
         return is df(index=contracts, colums=[mark_rpice, mark_vol,...])
         """
        if timestamp in self.chain_ts_group_by_time.groups:
            df = self.chain_ts_group_by_time.get_group(timestamp)
            df = df.set_index(SliceColumn.CONTRACT.value, drop=False)
            df.index.name = 'contract'

            if contracts is not None:
                df = df.iloc[np.in1d(df.index, contracts, assume_unique=False), :]
        else:
            # print(f"no slice data for {timestamp}")
            df = pd.DataFrame()
        return df

    def get_spot_price(self,
                       value_time: pd.Timestamp = None,
                       index: Union[pd.DatetimeIndex, pd.Index] = None
                       ) -> Union[float, pd.Series]:
        value = self.spot_data['close'].rename(self.ticker)
        if value_time is not None:
            if value_time in value.index:
                value = value[value_time]
            else:
                raise KeyError(f"in get_spot_price {value_time} not in {value.index}")
        elif index is not None:
            value = value.reindex(index=index, method='ffill').fillna(method='ffill')
        return value

    def get_funding_rate(self,
                         freq: Optional[Literal['H', 'D']] = 'H',
                         index: Union[pd.DatetimeIndex, pd.Index] = None,
                         is_rescale_to_one_hour: bool = True,
                         time_period: TimePeriod = None
                         ) -> pd.Series:
        """
        deribit reports hourly funding rate extrapolated to 8h interval
        for h freq we resample at 8h and return last
        for d freq we resample at 8h and then return the sum
        if freq is Non we return actual rate, it needs to be divided by 8 for comps H or D frequences with is_rescale_to_one_hour = True
        """
        funding_rate_1h = self.spot_data['funding_rate'].rename(self.ticker)
        if freq is not None:
            funding_rate_8h = funding_rate_1h.resample('8h').last()
            if freq == 'H':
                funding_rate = funding_rate_8h
                if index is not None:
                    funding_rate = funding_rate.reindex(index=index).fillna(0.0)
            elif freq == 'D':
                funding_rate = funding_rate_8h.resample('D').sum()
                if index is not None:
                    funding_rate = funding_rate.reindex(index=index, method='ffill').fillna(0.0)
            else:
                raise NotImplementedError(f"freq={freq}")
        else:
            # funding_rate = funding_rate_1h.cumsum(0).reindex(index=index, method='ffill').diff(1)
            funding_rate = funding_rate_1h
            if is_rescale_to_one_hour:
                funding_rate = funding_rate / 8.0  # deribit computes funding based on 8h frequency

        if time_period is not None:
            funding_rate = time_period.locate(funding_rate)

        return funding_rate

    def get_ewm_vol(self,
                    value_time: pd.Timestamp,
                    span: float = 168  # =7*24
                    ) -> float:
        if self.ewm_vol is None:
            returns = np.log(self.spot_data['close']).diff()
            self.ewm_vol = qis.compute_ewm_vol(data=returns, span=span, af=24.0 * 365.0)
        idx = self.ewm_vol.index.get_indexer([value_time], method='ffill')
        vol = self.ewm_vol.iloc[idx].to_numpy()[0]
        return vol

    @classmethod
    def reduce(cls, obj: ChainTs, contracts: List[str]) -> ChainTs:
        """
        reduce chain to contain only given contracts
        """
        chain_ts = obj.chain_ts
        return ChainTs(chain_ts=chain_ts.loc[chain_ts[SliceColumn.CONTRACT.value].isin(contracts), :],
                       spot_data=obj.spot_data,
                       ticker=obj.ticker)


@dataclass
class FuturesChainTs(ChainTs):
    """
    implementation of options time series data
    """

    future_ticker_config: FutureTickerConfig = field(default=FutureTickerConfig.CMS, init=True)

    class FuturesDataColumns(str, Enum):
        MARK_PRICE = 'mark_price'
        OPEN = 'open'
        HIGH = 'high'
        LOW = 'low'
        CLOSE = 'close'
        VOLUME = 'usd_volume'
        VOLUME_CONTRACTS = 'contract_count'
        OI_VOLUME = 'oi_value_usd'
        OI_CONTRACTS = 'oi_contract_count'

    def get_term_structure(self,
                           value_time: pd.Timestamp,
                           data: str = 'mark_price'
                           ) -> pd.DataFrame:

        contracts_data = self.get_contracts_data(value_time=value_time, data=data).dropna().rename(data).to_frame()
        ttms = [get_ttm_from_future_ticker(value_time=value_time, contract=x) for x in contracts_data.index]
        contracts_data['ttms'] = ttms
        return contracts_data


@dataclass
class OptionsDataDFs(ChainTs):
    """
    implementation of options time series data
    fields of SliceColumn are contained in data_dict
    """
    # do not pass to ChainTs
    # todo: do you need it here if all info (maturity, option type, etc is passed inside option data dfs)
    #option_ticker_config: OptionTickerConfig = field(default=OptionTickerConfig.DERIBIT, init=True)

    @classmethod
    def reduce(cls, obj: OptionsDataDFs, contracts: List[str]) -> OptionsDataDFs:
        """
        reduce chain to contain only given contracts
        """
        chain_ts = obj.chain_ts
        return OptionsDataDFs(chain_ts=chain_ts.loc[chain_ts[SliceColumn.CONTRACT.value].isin(contracts), :],
                              spot_data=obj.spot_data,
                              ticker=obj.ticker)
                              #option_ticker_config=obj.option_ticker_config)


class UnitTests(Enum):
    TEST_CHAIN = 1


def run_unit_test(unit_test: UnitTests):

    import local_path as local_path

    if unit_test == UnitTests.TEST_CHAIN:
        ticker = 'ETH'
        chain_ts = qis.load_df_from_feather(file_name=f"{ticker}_freq_H",
                                            index_col=None,
                                            local_path=f"{local_path.get_resource_path()}\\tardis\\")
        spot_data = qis.load_df_from_feather(file_name=f"{ticker}-spot",
                                             local_path=f"{local_path.get_resource_path()}\\deribit\\")

        chain_ts = ChainTs(chain_ts=chain_ts, spot_data=spot_data, ticker=ticker)
        chain_ts.print()
        chain_ts.get_start_end_date().print()

        df = chain_ts.get_time_slice(timestamp=pd.Timestamp('2019-04-01 08:00:00+00:00'))
        print(df)


if __name__ == '__main__':

    unit_test = UnitTests.TEST_CHAIN

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Tuple, Optional
from enum import Enum
import qis

from option_chain_analytics.option_chain import ExpirySlice, SliceColumn
from option_chain_analytics.visuals.utils import map_deltas_to_str


def plot_slice_open_interest(eslice: ExpirySlice,
                             title: str = None,
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> None:
    df = eslice.get_slice_open_interest()
    kwargs = qis.update_kwargs(kwargs,
                               dict(x_rotation=90, ncol=2, legend_loc='upper center', first_color_fixed=True))
    qis.plot_bars(df=df,
                  stacked=True,
                  var_format='{:,.0f}',
                  yvar_format='{:,.0f}',
                  is_top_totals=False,
                  # add_bar_values=True,
                  title=title,
                  ax=ax,
                  **kwargs)


def plot_slice_vols(eslice: ExpirySlice,
                    is_delta_space: bool = False,
                    delta_bounds: Tuple[Optional[float], Optional[float]] = None,
                    is_filtered: bool = False,
                    title: str = None,
                    delta_str_format: str = '0.2f',
                    ax: plt.Subplot = None,
                    **kwargs
                    ) -> None:

    df, strikes = eslice.get_bid_mark_ask_vols(is_delta_space=is_delta_space, delta_bounds=delta_bounds, is_filtered=is_filtered)
    if is_delta_space:
        max_delta = np.max(df.index.to_numpy())  # ATM is max absolute delta
        label_x_y = {'ATM strike': (f"{max_delta:{delta_str_format}}", df.loc[max_delta, SliceColumn.MARK_IV])}
        df.index = map_deltas_to_str(df.index.to_numpy(), delta_str_format=delta_str_format)
        xlabel, xvar_format, x_rotation = 'Delta', None, 90
    else:
        atm_strike = eslice.get_atm_option_strike()
        label_x_y = {'ATM strike': (atm_strike, df.loc[atm_strike, SliceColumn.MARK_IV])}
        xlabel, xvar_format,  x_rotation = 'Strike', '{:,.0f}', 0

    kwargs = qis.update_kwargs(kwargs,
                               dict(ncol=3, legend_loc='upper center',
                                    xvar_format=xvar_format,
                                    yvar_format='{:,.2%}',
                                    markersize=10,
                                    x_rotation=x_rotation))
    if ax is None:
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(6, 6), tight_layout=True)

    qis.plot_line(df=df,
                  linestyles=['', '-', ''],
                  markers=["^", "o", "v"],
                  ylabel='Implied Vol',
                  xlabel=xlabel,
                  title=title,
                  ax=ax,
                  **kwargs)
    qis.add_scatter_points(ax=ax, label_x_y=label_x_y, linewidth=10)


def plot_slice_vols_with_oi(eslice: ExpirySlice,
                            delta_bounds: Tuple[Optional[float], Optional[float]] = None,
                            is_filtered: bool = False,
                            is_delta_space: bool = False,
                            title: str = None,
                            delta_str_format: str = '0.2f',
                            ax: plt.Subplot = None,
                            **kwargs
                            ) -> None:
    if ax is None:
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(6, 9), tight_layout=True)

    plot_slice_vols(eslice=eslice,
                    is_delta_space=is_delta_space,
                    delta_bounds=delta_bounds,
                    is_filtered=is_filtered,
                    delta_str_format=delta_str_format,
                    title=title,
                    ax=ax,
                    **kwargs)

    ymin, ymax = ax.get_ylim()
    if ymin < 0.0:
        ax.set_ylim([-0.01, None]) # slices with zero bids

    # add oi bars
    ax = ax.twinx()
    df = eslice.get_slice_open_interest()
    # need to reindex to account for delta_bounds=delta_bounds, is_filtered=is_filtered
    df1, strikes = eslice.get_bid_mark_ask_vols(is_delta_space=is_delta_space, delta_bounds=delta_bounds,
                                                is_filtered=is_filtered)
    df = df.reindex(index=strikes.to_numpy())
    if is_delta_space:
        df.index = map_deltas_to_str(df.index.to_numpy(), delta_str_format=delta_str_format)  # should match vols
        x_vals = np.arange(len(df.index))
        widths = 0.25
    else:
        x_vals = df.index.to_numpy()
        widths = np.minimum(np.min(np.abs(x_vals[1:] - x_vals[:-1])), 100.0)
    oi_sets = {'Puts OI': 'steelblue', 'Calls OI': 'turquoise'}
    for idx, (key, color) in enumerate(oi_sets.items()):
        ax.bar(x_vals, df.iloc[:, idx].to_numpy(), widths, color=color, alpha=0.5, label=key, edgecolor='none', linewidth=0)
    qis.set_legend(ax=ax,
                   labels=list(oi_sets.keys()),
                   colors=list(oi_sets.values()),
                   bbox_to_anchor=(0.4, 0.95),
                   ncol=2,
                   **kwargs)
    qis.set_ax_xy_labels(ax=ax, ylabel='Contracts OI', **kwargs)
    ax.set_ylim([0.0, None])


class UnitTests(Enum):
    PRINT_CHAIN_DATA = 1
    PLOT_SLICE_OI = 2
    PLOT_SLICE_VOL = 3
    PLOT_SLICE_VOL_OI = 4


def run_unit_test(unit_test: UnitTests):

    from option_chain_analytics.ts_loaders import ts_data_loader_wrapper, DataSource
    from option_chain_analytics.data.chain_loader_from_ts import create_chain_from_from_options_dfs
    from option_chain_analytics.data.chain_ts import OptionsDataDFs

    pd.set_option('display.max_rows', 500)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.width', 1000)

    ticker = 'ETH'
    value_time = pd.Timestamp('2023-02-07 08:00:00+00:00')
    slice_id = '31MAR23'

    options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker, data_source=DataSource.TARDIS_LOCAL))
    chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)
    chain.print_slices_id()

    if unit_test == UnitTests.PRINT_CHAIN_DATA:
        for expiry, eslice in chain.expiry_slices.items():
            eslice.print()

    elif unit_test == UnitTests.PLOT_SLICE_OI:
        eslice = chain.expiry_slices[slice_id]
        plot_slice_open_interest(eslice=eslice)

    elif unit_test == UnitTests.PLOT_SLICE_VOL:
        eslice = chain.expiry_slices[slice_id]
        plot_slice_vols(eslice=eslice)
        plot_slice_vols(eslice=eslice, is_delta_space=True)

    elif unit_test == UnitTests.PLOT_SLICE_VOL_OI:
        eslice = chain.expiry_slices[slice_id]
        plot_slice_vols_with_oi(eslice=eslice, is_delta_space=False, delta_bounds=(-0.1, 0.1))
        #plot_slice_vols_with_oi(eslice=eslice, is_delta_space=True)

    mng = plt.get_current_fig_manager()
    mng.window.showMaximized()
    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_SLICE_VOL

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import numpy as np
from scipy.stats import t
from scipy.special import gamma
from typing import Union


def t_cum(y: float, nu: float) -> float:
    c = (1.0/np.sqrt(np.pi*nu))* (nu/(nu-1.0)) * gamma(0.5*(nu+1.0)) / gamma(0.5*nu)
    f = np.power(1.0+np.square(y) / nu, -0.5*(nu-1.0))
    return c * f


def compute_tdist_price(forward: Union[float, np.ndarray],
                        strikes: Union[float, np.ndarray],
                        ttm: float,
                        vol: float,
                        nu: float = 4.5,
                        optiontypes: Union[str, np.ndarray] = 'C',
                        discfactor: float = 1.0
                        ) -> Union[float, np.ndarray]:
    """
    bsm pricer for forward
    """
    scaler = vol*np.sqrt(ttm)*np.sqrt(0.5*nu)*((nu-1.0)/nu)*gamma(0.5*nu)/gamma(0.5*(nu+1.0))
    ups = 1.0 / scaler

    def compute(strike_: Union[float, np.ndarray], optiontype_: str) -> float:
        y = strike_ / forward - 1.0
        if optiontype_ == 'C' or optiontype_ == 'IC':
            price_ = discfactor * (forward * t_cum(ups * y, nu=nu) / ups + (forward - strike_) * (1.0 - t.cdf(ups * y, nu)))
        elif optiontype_ == 'P' or optiontype_ == 'IP':
            price_ = discfactor * (forward * t_cum(ups * y, nu=nu) / ups - (forward - strike_) * t.cdf(ups * y, nu))
        else:
            raise NotImplementedError(f"optiontype")
        return price_

    if isinstance(optiontypes, str):
        price = compute(strikes, optiontypes)
    else:
        price = np.zeros_like(strikes)
        for idx, (strike_, optiontype_) in enumerate(zip(strikes, optiontypes)):
            price[idx] = compute(strike_, optiontype_)
    return price


def compute_compute_negative_prob(ttms: np.ndarray,
                                  vol: float,
                                  drift: float = 0.0,
                                  nu: float = 3.5
                                  ) -> Union[float, np.ndarray]:
    """
    bsm pricer for forward
    """
    probs = np.zeros_like(ttms)
    q = vol*np.sqrt(0.5*nu)*((nu-1.0)/nu)*gamma(0.5*nu)/gamma(0.5*(nu+1.0))
    for idx, ttm in enumerate(ttms):
        scaler = q*np.sqrt(ttm)
        ups = 1.0 / scaler
        probs[idx] = t.cdf(-(1.0+drift*ttm)*ups, nu)
    return probs


def infer_tdist_implied_vol(forward: float,
                            ttm: float,
                            strike: float,
                            given_price: float,
                            discfactor: float = 1.0,
                            optiontype: str = 'C',
                            nu: float = 4.5,
                            tol: float = 1e-12,
                            is_bounds_to_nan: bool = False
                            ) -> float:
    """
    compute normal implied vol
    """
    x1, x2 = 0.05, 10.0  # starting values
    f = compute_tdist_price(forward=forward, strikes=strike, ttm=ttm, vol=x1, nu=nu, discfactor=discfactor, optiontypes=optiontype) - given_price
    fmid = compute_tdist_price(forward=forward, strikes=strike, ttm=ttm, vol=x2, nu=nu, discfactor=discfactor, optiontypes=optiontype) - given_price
    if f*fmid < 0.0:
        if f < 0.0:
            rtb = x1
            dx = x2-x1
        else:
            rtb = x2
            dx = x1-x2
        xmid = rtb
        for j in range(0, 100):
            dx = dx*0.5
            xmid = rtb+dx
            fmid = compute_tdist_price(forward=forward, strikes=strike, ttm=ttm, vol=xmid, nu=nu, discfactor=discfactor, optiontypes=optiontype) - given_price
            if fmid <= 0.0:
                rtb = xmid
            if np.abs(fmid) < tol:
                break
        v1 = xmid
    else:
        if f < 0:
            v1 = x1
        else:
            v1 = x2
    if is_bounds_to_nan:  # in case vol was inferred it will return nan
        if np.abs(v1-x1) < tol or np.abs(v1-x2) < tol:
            v1 = np.nan
    return v1


def infer_tdist_implied_vols_from_model_slice_prices(ttm: float,
                                                     forward: float,
                                                     strikes: np.ndarray,
                                                     optiontypes: np.ndarray,
                                                     model_prices: np.ndarray,
                                                     discfactor: float,
                                                     nu: float
                                                     ) -> np.ndarray:
    model_vol_ttm = np.zeros_like(strikes)
    for idx, (strike, model_price, optiontype) in enumerate(zip(strikes, model_prices, optiontypes)):
        model_vol_ttm[idx] = infer_tdist_implied_vol(forward=forward, ttm=ttm, discfactor=discfactor,
                                                     given_price=model_price,
                                                     strike=strike,
                                                     optiontype=optiontype,
                                                     nu=nu)
    return model_vol_ttm




"""
add prop data loaders here
the return type is Dict[str, Union[Dict[str, pd.DataFrame], pd.DataFrame]]: where str is SliceColumn
"""
from typing import Dict, Any
from enum import Enum

# public data api
from option_chain_analytics.public_apis.deribit import load_deribit_contract_ts_data

# prop data api
from prop.tardis.deribit_local import load_contract_ts_data_tardis


class DataSource(Enum):
    TARDIS_LOCAL = 1
    DERIBIT_LOCAL = 2


def ts_data_loader_wrapper(data_source: DataSource = DataSource.TARDIS_LOCAL,
                           ticker: str = 'BTC',
                           **kwargs
                           ) -> Dict[str, Any]:
    """
    generic wrapper for loading
    """
    if data_source == DataSource.TARDIS_LOCAL:
        return load_contract_ts_data_tardis(ticker=ticker, **kwargs)

    elif data_source == DataSource.DERIBIT_LOCAL:
        return load_deribit_contract_ts_data(ticker=ticker, **kwargs)

    else:
        raise NotImplementedError(f"{data_source}")



"""
iinterface for gen vol gen vol slices data
contract convention is 'deribit-SOL-28OCT22-10-C-option'
"""
from __future__ import annotations  # to allow class method annotations

import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from dataclasses import dataclass
from typing import Union, Dict, Tuple, Optional, Literal
from numba.typed import List
from enum import Enum
import qis

import local_path
# analytics
import option_chain_analytics.pricers.bsm as bsm
from option_chain_analytics.data.config import NearestStrikeOnGrid, StrikeSelection, compute_time_to_maturity

IS_FORCE_ALL_COLUMNS_DATA = False


class UnderlyingColumn(str, Enum):
    """
    mandatory field for underlying data
    """
    EXPIRY_ID = 'expiry_id'
    VALUE_TIME = 'value_time'
    EXPIRY = 'expiry'
    SPOT_PRICE = 'spot_price'
    UNDERLYING_INDEX = 'underlying_index'
    FORWARD_PRICE = 'forward_price'
    IR_RATE = 'ir_rate'
    TTM = 'ttm'


class SliceColumn(str, Enum):
    """
    mandatory columns for calls and puts dataframes
    for crypto inverse options, mark_price is in units of underlying and usd_multiplier = underlying_price
    """
    # comes from data set
    CONTRACT = 'contract'  # string
    EXCHANGE_TIME = 'exchange_time'  # pd.Timestamp  record time of quate at exchange
    UNDERLYING_INDEX = 'underlying_index'  # id of underlying asset
    UNDERLYING_PRICE = 'underlying_price'  # price of underlying asset
    USD_MULTIPLIER = 'usd_multiplier'  # usd_multiplier = underlying_price or 1.0
    MARK_PRICE = 'mark_price'  # option mark
    BID_PRICE = 'bid_price'  # best bid
    ASK_PRICE = 'ask_price'   # best ask
    BID_SIZE = 'bid_size'
    ASK_SIZE = 'ask_size'
    MARK_IV = 'mark_iv'  # float
    BID_IV = 'bid_iv'  # float
    ASK_IV = 'ask_iv'  # float
    DELTA = 'delta'  # float
    VEGA = 'vega'  # float
    THETA = 'theta'  # float
    GAMMA = 'gamma'  # float
    OPEN_INTEREST = 'open_interest'  # number of open contratcs,  float
    VOLUME = 'volume'  # float
    MATURITY_ID = 'mat_id'  # str maturity id
    STRIKE = 'strike'  # float
    OPTION_TYPE = 'optiontype'  # {'C', 'P'}
    EXPIRY = 'expiry'  # pd.Timestamp
    TTM = 'ttm'  # float
    CONTRACT_SIZE = 'contract_size'  # float
    INTEREST_RATE = 'interest_rate'  # float


def get_clean_slice(df: pd.DataFrame) -> pd.DataFrame:
    cond = (df[SliceColumn.MARK_PRICE] > 0.0) \
           & (df[SliceColumn.BID_PRICE] > 0.0) & (df[SliceColumn.ASK_PRICE] > 0.0)  \
           & (df[SliceColumn.UNDERLYING_PRICE] > 0.0)
    cond = (df[SliceColumn.MARK_PRICE] > 0.0)
    vol_cond = (df[SliceColumn.BID_IV].isna() == False) & (df[SliceColumn.ASK_IV].isna() == False)\
               & (df[SliceColumn.MARK_IV].isna() == False)
    clean_df = df[cond]
    return clean_df


@dataclass
class ExpirySlice:
    """
    slice_t of call and puts options data per expiry
    """
    undelying_data: pd.Series
    options_df: pd.DataFrame

    def __post_init__(self):

        if IS_FORCE_ALL_COLUMNS_DATA:
            ALL_COLUMNS = list(x.value for x in SliceColumn)
            all_options_col = np.all(np.in1d(self.options_df.columns, ALL_COLUMNS, assume_unique=True))
            if not all_options_col:
                raise ValueError(f"missing column data "
                                 f"{self.options_df.columns[np.in1d(self.options_df.columns, ALL_COLUMNS, assume_unique=True)]}")

        self.expiry_id = self.undelying_data[UnderlyingColumn.EXPIRY_ID]
        self.value_time = self.undelying_data[UnderlyingColumn.VALUE_TIME]
        self.expiry_time = self.undelying_data[UnderlyingColumn.EXPIRY]
        self.forward = self.undelying_data[UnderlyingColumn.FORWARD_PRICE]
        self.ttm = self.undelying_data[UnderlyingColumn.TTM]
        # for easy of searching contracts by strike, reindex call and puts using strikes
        self.calls = self.options_df.loc[self.options_df[SliceColumn.OPTION_TYPE.value] == 'C', :].sort_values(
            by=SliceColumn.STRIKE).set_index(SliceColumn.STRIKE, drop=False)
        self.puts = self.options_df.loc[self.options_df[SliceColumn.OPTION_TYPE.value] == 'P', :].sort_values(
            by=SliceColumn.STRIKE).set_index(SliceColumn.STRIKE, drop=False)

    def print(self) -> None:
        print(f"expiry_id={self.expiry_id},\n"
              f"future_price={self.forward},\n"
              f"value_time={self.value_time},\n"
              f"expiry_time={self.expiry_time}\n")
        print(f"calls=\n{self.calls}")
        print(f"puts=\n{self.puts}")
    
    def get_options_df(self, contracts: Optional[List[str]] = None) -> pd.DataFrame:
        options_df = self.options_df.copy()
        options_df = options_df.set_index(SliceColumn.CONTRACT, drop=False)
        if contracts is not None:
            options_df = options_df.loc[contracts, :]
        return options_df

    def get_future_price(self) -> float:
        return self.forward

    def get_ttm(self) -> float:
        return self.ttm

    def get_usd_mark(self) -> pd.Series:
        """
        for coined options mark_price is in BTC/ETH, multiply by underlying future price
        """
        return self.options_df[SliceColumn.MARK_PRICE].multiply(self.options_df[SliceColumn.USD_MULTIPLIER])

    def get_usd_ask(self) -> pd.Series:
        """
        for coined options mark_price is in BTC/ETH, multiply by underlying future price
        """
        return self.options_df[SliceColumn.ASK_PRICE].multiply(self.options_df[SliceColumn.USD_MULTIPLIER])

    def get_mid_bid_ask(self) -> pd.Series:
        """
        mid_price = contract_data[SliceColumn.MARK_PRICE]
        we take mid if both bid and ask exists, otherwise we take mid
        """
        mark_price = self.options_df[SliceColumn.MARK_PRICE].to_numpy()
        bid_price = self.options_df[SliceColumn.BID_PRICE].to_numpy()
        ask_price = self.options_df[SliceColumn.ASK_PRICE].to_numpy()
        mid_price = pd.Series(np.where(np.logical_and(np.isnan(bid_price == False), np.isnan(ask_price == False)),
                                       0.5 * (bid_price + ask_price),
                                       mark_price),
                              index=self.options_df.index)
        return mid_price.multiply(self.options_df[SliceColumn.USD_MULTIPLIER])

    def get_call_slice(self,
                       is_filtered: bool = True,
                       min_call_strike: float = None,
                       max_call_strike: float = None,
                       index_by_contract: bool = False
                       ) -> pd.DataFrame:
        if is_filtered:
            df = get_clean_slice(df=self.calls)
        else:
            df = self.calls
        if min_call_strike is not None:
            df = df.loc[df[SliceColumn.STRIKE] >= min_call_strike, :]
        if max_call_strike is not None:
            df = df.loc[df[SliceColumn.STRIKE] <= max_call_strike, :]
        if df.empty:
            print(f"empty slice call data at {self.expiry_id} on {self.value_time}")
        if not df.empty and index_by_contract is True:
            df = df.set_index(SliceColumn.CONTRACT, drop=False)
        return df

    def get_put_slice(self,
                      is_filtered: bool = True,
                      index_by_contract: bool = False,
                      max_put_strike: Optional[float] = None,
                      min_put_strike: Optional[float] = None,
                      delta_floor: Optional[float] = None
                      ) -> pd.DataFrame:
        if is_filtered:
            df = get_clean_slice(df=self.puts)
        else:
            df = self.puts
        if max_put_strike is not None:
            df = df.loc[df[SliceColumn.STRIKE] <= max_put_strike, :]
        if min_put_strike is not None:
            df = df.loc[df[SliceColumn.STRIKE] >= min_put_strike, :]
        if delta_floor is not None:
            df = df.loc[df[SliceColumn.DELTA] <= delta_floor, :]
        if df.empty:
            print(f"empty slice put data at {self.expiry_id} on {self.value_time}")
        if not df.empty and index_by_contract is True:
            df = df.set_index(SliceColumn.CONTRACT, drop=False)
        return df

    def get_joint_slice(self,
                        index_by_contract: bool = False,
                        delta_bounds: Tuple[Optional[float], Optional[float]] = None,
                        is_filtered: bool = True,
                        min_put_strike: Optional[float] = None,
                        max_put_strike: Optional[float] = None,
                        max_call_strike: Optional[float] = None,
                        min_call_strike: Optional[float] = None,
                        forward_ref: float = None
                        ) -> pd.DataFrame:
        """
        get put/call ivols slice_t
        """
        put_wing = self.get_put_slice(index_by_contract=index_by_contract, is_filtered=is_filtered,
                                      max_put_strike=max_put_strike, min_put_strike=min_put_strike)
        call_wing = self.get_call_slice(index_by_contract=index_by_contract, is_filtered=is_filtered,
                                        min_call_strike=min_call_strike, max_call_strike=max_call_strike)
        if put_wing.empty or call_wing.empty:
            print(f"empty joint slice at {self.expiry_id} on {self.value_time}")
            joint_slice = pd.DataFrame()
        else:
            forward_ref = forward_ref or self.forward
            if max_put_strike is not None:  # the put wing cut is done inside get_put_slice
                put_wing = put_wing
            else:
                put_wing = put_wing.loc[put_wing[SliceColumn.STRIKE] <= forward_ref, :]

            if delta_bounds is not None and delta_bounds[0] is not None:
                put_wing = put_wing.loc[put_wing[SliceColumn.DELTA] <= delta_bounds[0], :]

            if min_call_strike is not None:     # the call wing cut is done inside get_call_slice
                call_wing = call_wing
            else:
                call_wing = call_wing.loc[call_wing[SliceColumn.STRIKE] >= forward_ref, :]

            if delta_bounds is not None and delta_bounds[1] is not None:
                call_wing = call_wing.loc[call_wing[SliceColumn.DELTA] >= delta_bounds[1], :]
            joint_slice = pd.concat([put_wing, call_wing], axis=0)  # stitch together
        return joint_slice

    def get_bid_mark_ask_vols(self,
                              is_delta_space: bool = False,
                              delta_bounds: Tuple[Optional[float], Optional[float]] = None,
                              is_filtered: bool = True,
                              drop_same_deltas: bool = True  # in delta space deep oot options end up with same delta
                              ) -> Tuple[Optional[pd.DataFrame], Optional[pd.Series]]:
        df = self.get_joint_slice(delta_bounds=delta_bounds, is_filtered=is_filtered)
        if not df.empty:
            if is_delta_space:
                df1 = df[[SliceColumn.DELTA, SliceColumn.BID_IV, SliceColumn.MARK_IV, SliceColumn.ASK_IV, SliceColumn.STRIKE]].set_index(
                    SliceColumn.DELTA)
                if drop_same_deltas:
                    df1.index = np.around(df1.index, decimals=4)
                    df1 = df1[~df1.index.duplicated(keep='first')]
            else:
                df1 = df[[SliceColumn.STRIKE, SliceColumn.BID_IV, SliceColumn.MARK_IV, SliceColumn.ASK_IV]].set_index(
                    SliceColumn.STRIKE, drop=False)

            strikes = df1[SliceColumn.STRIKE]
            df1 = df1.drop(SliceColumn.STRIKE, axis=1)
            return df1, strikes
        else:
            return None, None

    def get_vols_with_logstrikes(self,
                                 is_delta_space: bool = False,
                                 delta_bounds: Tuple[Optional[float], Optional[float]] = None,
                                 is_filtered: bool = True,
                                 vol_type: Literal[SliceColumn.BID_IV, SliceColumn.MARK_IV, SliceColumn.ASK_IV] = SliceColumn.ASK_IV
                                 ) -> Tuple[pd.Series, np.ndarray]:
        df1, strikes = self.get_bid_mark_ask_vols(is_delta_space=is_delta_space,
                                                  delta_bounds=delta_bounds,
                                                  is_filtered=is_filtered)
        vols = df1[vol_type]
        log_strikes = np.log(vols.index.to_numpy() / self.forward)
        return vols, log_strikes

    def get_atm_option_strike(self, nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI) -> Optional[float]:
        joint_slice = self.get_joint_slice(is_filtered=True)
        if not joint_slice.empty:
            a = joint_slice[SliceColumn.STRIKE].to_numpy()
            if nearest_strike_on_grid == NearestStrikeOnGrid.MAX_OI:
                weight = joint_slice[SliceColumn.OPEN_INTEREST].to_numpy()
            else:
                weight = None
            idx = find_idx_nearest_element(value=self.forward, a=a, weight=weight, nearest_strike_on_grid=nearest_strike_on_grid)
            return joint_slice.index[idx]
        else:
            return None

    def get_atm_vol(self, nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI) -> float:
        atm_strike = self.get_atm_option_strike(nearest_strike_on_grid=nearest_strike_on_grid)
        call_vol = self.calls.loc[atm_strike, SliceColumn.MARK_IV]
        put_vol = self.puts.loc[atm_strike, SliceColumn.MARK_IV]
        return 0.5*(call_vol+put_vol)

    def get_atm_call_id(self, nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI) -> Optional[str]:
        strike = self.get_atm_option_strike(nearest_strike_on_grid=nearest_strike_on_grid)
        if strike is not None:
            if strike not in self.calls.index:  # MAX_OI may not work sometimes for joint slice
                strike = self.get_atm_option_strike(nearest_strike_on_grid=NearestStrikeOnGrid.BELOW)
            return self.calls.loc[strike, SliceColumn.CONTRACT]
        else:
            return None

    def get_atm_put_id(self, nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI) -> Optional[str]:
        strike = self.get_atm_option_strike(nearest_strike_on_grid=nearest_strike_on_grid)
        if strike is not None:
            return self.puts.loc[strike, SliceColumn.CONTRACT]
        else:
            return None

    def get_put_delta_strike(self,
                             delta: float = -0.25,
                             nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI
                             ) -> float:
        put_wing = self.get_put_slice(is_filtered=True)
        a = put_wing[SliceColumn.DELTA].to_numpy()
        if nearest_strike_on_grid == NearestStrikeOnGrid.MAX_OI:
            weight = put_wing[SliceColumn.OPEN_INTEREST].to_numpy()
        else:
            weight = None
        idx = find_idx_nearest_element(value=delta, a=a, weight=weight, nearest_strike_on_grid=nearest_strike_on_grid)
        return put_wing.index[idx]

    def get_put_delta_option_id(self,
                                delta: float = -0.25,
                                nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI
                                ) -> str:
        strike = self.get_put_delta_strike(delta=delta, nearest_strike_on_grid=nearest_strike_on_grid)
        return self.puts.loc[strike, SliceColumn.CONTRACT]

    def get_call_delta_strike(self,
                              delta: float = 0.25,
                              nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI
                              ) -> float:
        call_wing = self.get_call_slice(is_filtered=True)
        a = call_wing[SliceColumn.DELTA].to_numpy()
        if nearest_strike_on_grid == NearestStrikeOnGrid.MAX_OI:
            weight = call_wing[SliceColumn.OPEN_INTEREST].to_numpy()
        else:
            weight = None
        idx = find_idx_nearest_element(value=delta, a=a, weight=weight, nearest_strike_on_grid=nearest_strike_on_grid)
        return call_wing.index[idx]

    def get_call_delta_option_id(self, delta: float = 0.25,
                                 nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI
                                 ) -> str:
        strike = self.get_call_delta_strike(delta=delta, nearest_strike_on_grid=nearest_strike_on_grid)
        return self.calls.loc[strike, SliceColumn.CONTRACT]

    def get_slice_open_interest(self) -> pd.DataFrame:
        calls = self.calls[SliceColumn.OPEN_INTEREST].rename('C')
        puts = self.puts[SliceColumn.OPEN_INTEREST].rename('P')
        df = pd.concat([puts, calls], axis=1)
        return df


@dataclass
class SlicesChain:
    """
    collections of slices in time
    """
    options_df: pd.DataFrame   # all options data index = all contracts, columns = SliceColumn
    undelying_df: pd.DataFrame   # spot, rate ttm index= slice ids, columns = UnderlyingColumn
    expiry_slices: Dict[str, ExpirySlice]  # arranged expiry slices, equivalent to options_df.groupby(expiry_id)
    value_time: pd.Timestamp  # valu_time of the chain

    def __post_init__(self):
        # sort by maturities
        mats = {key: eslice.get_ttm() for key, eslice in self.expiry_slices.items()}
        sorted_dds_dict = dict(sorted(mats.items(), key=lambda item: item[1]))
        self.expiry_slices = {k: self.expiry_slices[k] for k in sorted_dds_dict.keys()}  # sorded {Slice_id: ExpirySlice}
        self.expiry_maturities = {k: v.get_ttm() for k, v in self.expiry_slices.items()}  # {Slice_id: ttm}
        self.expiry_times = {v.expiry_time: k for k, v in self.expiry_slices.items()}  # {expiry time: slice id}  # useful to find sices by maturity dates

    def print_slices_id(self) -> None:
        for key in self.expiry_slices.keys():
            print(key)

    def get_expiry_slice(self, slice_id: str) -> ExpirySlice:
        if slice_id in self.expiry_slices.keys():
            return self.expiry_slices[slice_id]
        else:
            raise KeyError(f"{slice_id} not in {self.expiry_slices.keys()}")

    def get_slice_id_for_mat_date(self, mat_date: pd.Timestamp, is_normalized: bool = False) -> Optional[str]:
        if is_normalized:  # can find maturity without hour and tz info
            mat_date = mat_date.normalize().tz_convert(None)
            expiry_times = {k.normalize().tz_convert(None): v for k, v in self.expiry_times.items()}  # {expiry time: slice id}  # useful to find sices by maturity dates
        else:
            expiry_times = self.expiry_times
        if mat_date in expiry_times.keys():
            return expiry_times[mat_date]
        else:
            return None

    def get_slice_id_for_mat_dates(self, mat_dates: Union[List[pd.Timestamp], pd.DatetimeIndex],
                                   is_apply_open_interest_filter: bool = False,
                                   is_normalized: bool = True  # does not depend on expiry date
                                   ) -> List[str]:
        slice_ids = []
        for mat_date in mat_dates:
            slice_id = self.get_slice_id_for_mat_date(mat_date=mat_date, is_normalized=is_normalized)
            # by default remove slices with less than 1000 - typical for just rolled contracts
            if slice_id is not None and is_apply_open_interest_filter:
                slice_t = self.expiry_slices[slice_id].get_joint_slice(is_filtered=True)
                if slice_t.empty:
                    slice_id = None
                else:
                    if np.nansum(slice_t[SliceColumn.OPEN_INTEREST]) < 1000.0:
                        slice_id = None
            if slice_id is not None:
                slice_ids.append(slice_id)
        return slice_ids

    def get_next_slice_after_date(self, mat_date: pd.Timestamp) -> str:
        ttm = compute_time_to_maturity(maturity_time=mat_date, value_time=self.value_time)
        a = np.array(list(self.expiry_maturities.values()))
        idx = np.searchsorted(a=a, v=ttm, side='left')
        return list(self.expiry_maturities.keys())[idx]

    def get_atm_call_id(self,
                        mat_date: pd.Timestamp = None,
                        slice_id: str = None,
                        nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI
                        ) -> str:
        if slice_id is None and mat_date is None:
            raise ValueError(f"provide slice_id or mat_date")
        if slice_id is None:
            slice_id = self.get_slice_id_for_mat_date(mat_date)
        slice_t = self.expiry_slices[slice_id]
        return slice_t.get_atm_call_id(nearest_strike_on_grid=nearest_strike_on_grid)

    def get_atm_put_id(self,
                       mat_date: pd.Timestamp = None,
                       slice_id: str = None,
                       nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI
                       ) -> str:
        if slice_id is None and mat_date is None:
            raise ValueError(f"provide slice_id or mat_date")
        if slice_id is None:
            slice_id = self.get_slice_id_for_mat_date(mat_date)
        slice_t = self.expiry_slices[slice_id]
        return slice_t.get_atm_put_id(nearest_strike_on_grid=nearest_strike_on_grid)

    def get_atm_vol(self,
                    mat_date: pd.Timestamp = None,
                    slice_id: str = None,
                    nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI
                    ) -> Optional[float]:
        atm_call_id = self.get_atm_call_id(mat_date=mat_date, slice_id=slice_id, nearest_strike_on_grid=nearest_strike_on_grid)
        atm_put_id = self.get_atm_put_id(mat_date=mat_date, slice_id=slice_id, nearest_strike_on_grid=nearest_strike_on_grid)

        call_vol = None  # only if both bid and ask are available
        if atm_call_id is not None:
            call = self.get_contract_data(contract=atm_call_id)
            call_vol = call[SliceColumn.MARK_IV]

        put_vol = None # only if both bid and ask are available
        if atm_put_id is not None:
            put = self.get_contract_data(contract=atm_put_id)
            put_vol = put[SliceColumn.MARK_IV]

        if call_vol is not None and put_vol is not None:
            atm_vol = np.nanmean(np.array([call_vol, put_vol]))
        else:
            if call_vol is not None:
                atm_vol = call_vol
            elif put_vol is not None:
                atm_vol = put_vol
            else:
                atm_vol = None

        return atm_vol

    def get_put_delta_option_id(self,
                                mat_date: pd.Timestamp = None,
                                slice_id: str = None,
                                delta: float = -0.25,
                                nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI
                                ) -> str:
        if slice_id is None and mat_date is None:
            raise ValueError(f"provide slice_id or mat_date")
        if slice_id is None:
            slice_id = self.get_slice_id_for_mat_date(mat_date)
        return self.expiry_slices[slice_id].get_put_delta_option_id(delta=delta, nearest_strike_on_grid=nearest_strike_on_grid)

    def get_call_delta_option_id(self, mat_date: pd.Timestamp = None, slice_id: str = None, delta: float = 0.25,
                                 nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI
                                 ) -> str:
        if slice_id is None and mat_date is None:
            raise ValueError(f"provide slice_id or mat_date")
        if slice_id is None:
            slice_id = self.get_slice_id_for_mat_date(mat_date)
        return self.expiry_slices[slice_id].get_call_delta_option_id(delta=delta, nearest_strike_on_grid=nearest_strike_on_grid)

    def get_contract_data(self, contract: str) -> pd.Series:
        contract_data = self.options_df.loc[contract, :]
        return contract_data

    def get_contract_execution_price(self,
                                     contract: str,
                                     size: float,
                                     is_trade_at_bid_ask: bool = True,
                                     is_usd: bool = False
                                     ) -> Tuple[float, float]:
        contract_data = self.options_df.loc[contract, :]
        mid_price = contract_data[SliceColumn.MARK_PRICE]
        if is_trade_at_bid_ask:
            if size > 0.0:
                execution_price = contract_data[SliceColumn.ASK_PRICE]
            else:
                execution_price = contract_data[SliceColumn.BID_PRICE]
        else:
            execution_price = mid_price
        slippage = np.abs(execution_price-mid_price)

        if is_usd:
            execution_price *= contract_data[SliceColumn.USD_MULTIPLIER]
            slippage *= contract_data[SliceColumn.USD_MULTIPLIER]

        return execution_price, slippage

    def generate_delta_vol_matrix(self,
                                  value_time: pd.Timestamp,
                                  days_map: Dict[str, int] = {'1w': 7, '1m': 30},
                                  deltas: List[float] = (-0.10, -0.25, 0.50, 0.25, 0.10),
                                  nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.ABOVE,
                                  ) -> Optional[DeltaVolMatrix]:
        """
        generate table snapshot of vol data: index= days, columns = deltas
        """
        strikes_matrix, option_ids_matrix, vols_matrix, deltas_matrix, underlying_prices_matrix = {}, {}, {}, {}, {}
        labels = {}
        for label, day in days_map.items():
            next_date = value_time + pd.DateOffset(days=day)   # if overlapping next date will be last avilable maturity
            slice_date = self.get_next_slice_after_date(mat_date=next_date)
            slice_t = self.expiry_slices[slice_date]

            df = slice_t.get_joint_slice(is_filtered=True, delta_bounds=(-0.1, 0.1))

            if len(df.index) >= len(deltas):  # filter out slices with no data
                labels[label] = pd.Series(label, name=next_date)
                vols, option_ids, strikes, strike_deltas, index_prices = [], [], [], [], []
                for delta in deltas:
                    if delta < 0.0:
                        strike = slice_t.get_put_delta_strike(delta=delta, nearest_strike_on_grid=nearest_strike_on_grid)
                        slice_df = slice_t.get_put_slice()
                    else:
                        strike = slice_t.get_call_delta_strike(delta=delta, nearest_strike_on_grid=nearest_strike_on_grid)
                        slice_df = slice_t.get_call_slice()
                    vols.append(slice_df.loc[strike, SliceColumn.MARK_IV])
                    option_ids.append(slice_df.loc[strike, SliceColumn.CONTRACT])
                    strikes.append(strike)
                    strike_deltas.append(slice_df.loc[strike, SliceColumn.DELTA])
                    index_prices.append(slice_df.loc[strike, SliceColumn.UNDERLYING_PRICE])

                vols_matrix[label] = pd.Series(vols, index=deltas, name=next_date)
                strikes_matrix[label] = pd.Series(strikes, index=deltas, name=next_date)
                option_ids_matrix[label] = pd.Series(option_ids, index=deltas, name=next_date)
                deltas_matrix[label] = pd.Series(strike_deltas, index=deltas, name=next_date)
                underlying_prices_matrix[label] = pd.Series(index_prices, index=deltas, name=next_date)
        vols_matrix = pd.DataFrame.from_dict(vols_matrix, orient='index')
        option_ids_matrix = pd.DataFrame.from_dict(option_ids_matrix, orient='index')
        strikes_matrix = pd.DataFrame.from_dict(strikes_matrix, orient='index')
        deltas_matrix = pd.DataFrame.from_dict(deltas_matrix, orient='index')
        underlying_prices_matrix = pd.DataFrame.from_dict(underlying_prices_matrix, orient='index')

        labels = pd.DataFrame.from_dict(labels, orient='index')
        if not labels.empty:
            labels = labels.rename({0: 'tenor'}, axis=1)
            labels['days'] = labels.iloc[:, 0].map(days_map)
            labels = labels.sort_values('days')

            delta_vol_matrix = DeltaVolMatrix(value_time=value_time,
                                              labels=labels,
                                              vols_matrix=vols_matrix,
                                              option_ids_matrix=option_ids_matrix,
                                              strikes_matrix=strikes_matrix,
                                              deltas_matrix=deltas_matrix,
                                              underlying_prices_matrix=underlying_prices_matrix)
        else:
            delta_vol_matrix = None
        return delta_vol_matrix

    def plot_total_oi_in_maturity(self,
                                  title: str = None,
                                  is_usd: bool = True,
                                  ax: plt.Subplot = None,
                                  **kwargs
                                  ) -> None:
        totals_by_mat = []
        for slice_id in self.expiry_slices.keys():
            eslice = self.expiry_slices[slice_id]
            df = eslice.get_slice_open_interest()
            if is_usd:
                df = df*eslice.forward
            totals_by_mat.append(df.sum(0).rename(slice_id))

        totals_by_mat = pd.concat(totals_by_mat, axis=1).T
        totals_by_mat_all = totals_by_mat.sum(axis=0)
        totals_by_mat = totals_by_mat.rename({'P': f'Puts total = {totals_by_mat_all[0]:,.0f}',
                                              'C': f'Calls total = {totals_by_mat_all[1]:,.0f}'},
                                             axis=1)

        qis.plot_bars(totals_by_mat,
                  stacked=True,
                  yvar_format='{:,.0f}',
                  colors=['orangered', 'green'],
                  legend_loc='upper center',
                  title=title,
                  x_rotation=90,
                  ax=ax,
                  **kwargs)

    def plot_total_oi_in_strikes(self,
                                 title: str = None,
                                 is_usd: bool = True,
                                 ax: plt.Subplot = None,
                                 **kwargs
                                 ) -> None:
        calls_by_strike = {}
        puts_by_strike = {}
        for slice_id in self.expiry_slices.keys():
            eslice = self.expiry_slices[slice_id]
            df = eslice.get_slice_open_interest()
            if is_usd:
                df = eslice.forward * df
            for idx, row in df.iterrows():
                if idx in calls_by_strike.keys():
                    calls_by_strike[idx] += row['C']
                    puts_by_strike[idx] += row['P']
                else:
                    calls_by_strike[idx] = row['C']
                    puts_by_strike[idx] = row['P']

        calls_by_strike = pd.Series(calls_by_strike, name='C')
        puts_by_strike = pd.Series(puts_by_strike, name='P')
        totals_by_strike = pd.concat([puts_by_strike, calls_by_strike], axis=1).sort_index()
        totals_by_mat_all = totals_by_strike.sum(axis=0)
        totals_by_strike = totals_by_strike.rename({'P': f'Puts total = {totals_by_mat_all[0]:,.0f}',
                                                    'C': f'Calls total = {totals_by_mat_all[1]:,.0f}'},
                                                   axis=1)
        qis.plot_bars(totals_by_strike,
                      stacked=True,
                      yvar_format='{:,.0f}',
                      colors=['orangered', 'green'],
                      legend_loc='upper center',
                      title=title,
                      x_rotation=90,
                      ax=ax,
                      **kwargs)

    def save_joint_slices(self, file_name: str) -> None:
        slice_dfs = {}
        for key, slite_t in self.expiry_slices.items():
            slice_dfs[key] = slite_t.get_joint_slice()
        qis.save_df_dict_to_feather(dfs=slice_dfs, file_name=file_name, local_path=local_path.get_output_path())


@dataclass
class DeltaVolMatrix:
    """
    collections of slices in time and delta
    """
    value_time: pd.Timestamp
    labels: pd.DataFrame  # day
    vols_matrix: pd.DataFrame
    option_ids_matrix: pd.DataFrame
    strikes_matrix: pd.DataFrame
    deltas_matrix: pd.DataFrame
    underlying_prices_matrix: pd.DataFrame

    def print(self):
        print(self.labels)
        print(self.vols_matrix)
        print(self.option_ids_matrix)
        print(self.strikes_matrix)
        print(self.deltas_matrix)
        print(self.underlying_prices_matrix)

    def get_melted_matrix(self,
                          mat_id: str = 'tenor',
                          is_tenor: bool = True,
                          is_add_tenor_to_delta_name: bool = True
                          ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        datas = {'vols': self.vols_matrix,
                 'options': self.option_ids_matrix,
                 'strikes': self.strikes_matrix,
                 'underlying_prices_matrix': self.underlying_prices_matrix}
        out = {}
        if is_tenor:
            xdata = self.labels[mat_id]
        else:
            xdata = pd.Series(self.labels.index, index=self.labels.index, name=mat_id)
        for key, df in datas.items():
            df1 = qis.melt_scatter_data_with_xdata(df=df,
                                                   xdata=xdata,
                                                   y_column=key,
                                                   hue_name='delta')
            if is_add_tenor_to_delta_name:
                df1.index = [f"{delta:0.2f}d_{day}" for day, delta in zip(df1[mat_id], df1['delta'])]
            else:
                df1.index = [f"{delta:0.2f}" for delta in df1['delta']]
            out[key] = df1
        return out['vols'], out['strikes'], out['options'], out['underlying_prices_matrix']

    def plot_vol_matrix_table(self, ax: plt.Subplot = None, **kwargs):

        # data_colors = put.compute_heatmap_colors(a=self.vols_matrix.to_numpy(), **kwargs)
        df = qis.df_to_str(df=self.vols_matrix, var_format='{:,.0%}')
        qis.plot_df_table(df=df,
                          first_column_width=None,
                          index_column_name='Tenor/\nDelta',
                          # data_colors=data_colors,
                          # heatmap_rows=[x for x in range(0, len(df.columns)-1)],
                          special_columns_colors=[(0, 'steelblue'), (int(len(df.columns)/2+1), 'lightblue')],
                          ax=ax,
                          **kwargs)

    def plot_vol_in_strike(self,
                           title: str = 'ETH implied vols at strikes for delta grid = {-0.10, -0.25, 0.50, 0.25, 0.10}',
                           ax: plt.Subplot = None,
                           **kwargs
                           ) -> None:
        vols, strikes, _, _ = self.get_melted_matrix(is_tenor=False)
        vols['strikes'] = strikes['strikes']
        vols['delta'] = vols['delta'].apply(lambda x: str(x))

        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(12, 8))

        colors = qis.get_cmap_colors(n=len(vols['tenor'].unique()))
        sns.lineplot(data=vols, x='strikes', y='vols', hue='tenor', style='tenor', palette=colors, markers=False,
                     dashes=False, ax=ax)

        markers = qis.get_n_markers(n=len(vols['delta'].unique()))
        sns.scatterplot(data=vols, x="strikes", y="vols", hue="delta", style="delta", markers=markers, ax=ax)

        qis.set_ax_ticks_format(ax=ax, xvar_format='{:,.0f}', yvar_format='{:,.0%}', **kwargs)
        qis.set_title(ax=ax, title=title, **kwargs)

        qis.set_legend(ax=ax, **kwargs)

        h, l = ax.get_legend_handles_labels()
        this = len(h)
        n_tenor = len(self.deltas_matrix.index)
        n_deltas = len(self.deltas_matrix.columns)
        # l1 = ax.legend(h[:int(len(h) / 2)], l[:int(len(l) / 2)], loc='upper left', title='Option Tenor', fontsize=14)
        # l2 = ax.legend(h[int(len(h) / 2):], l[int(len(l) / 2):], loc='upper right', title='Deltas', ncol=5, fontsize=14)
        l1 = ax.legend(h[:n_tenor], l[:n_tenor], loc='upper left', title='Option Tenor', fontsize=14)
        l2 = ax.legend(h[n_tenor:], l[n_tenor:], loc='upper right', title='Deltas', ncol=5, fontsize=14)
        ax.add_artist(l1)   # we need this because the 2nd call to legend() erases the first


def find_idx_nearest_element(value: float,
                             a: np.ndarray,
                             weight: np.ndarray = None,
                             nearest_strike_on_grid: NearestStrikeOnGrid = NearestStrikeOnGrid.MAX_OI
                             ) -> int:
    """
    select nearest strike from a grid  
    """
    idx_sorted = np.argsort(a)
    sorted_array = a[idx_sorted]
    if weight is not None:
        weight = weight[idx_sorted]

    idx = np.searchsorted(sorted_array, value, side="left")
    if idx >= len(a):
        idx_nearest = idx_sorted[len(a)-1]
    elif idx == 0:
        idx_nearest = idx_sorted[0]
    else:
        if nearest_strike_on_grid == NearestStrikeOnGrid.MAX_OI and weight is not None:  # use biggest weight
            if weight[idx-1] > weight[idx]:
                idx_nearest = idx_sorted[idx-1]
            else:
                idx_nearest = idx_sorted[idx]

        elif nearest_strike_on_grid == NearestStrikeOnGrid.BELOW:
            if sorted_array[idx-1] > sorted_array[idx]:
                idx_nearest = idx_sorted[idx - 1]
            else:
                idx_nearest = idx_sorted[idx]

        elif nearest_strike_on_grid == NearestStrikeOnGrid.ABOVE:
            if sorted_array[idx-1] < sorted_array[idx]:
                idx_nearest = idx_sorted[idx - 1]
            else:
                idx_nearest = idx_sorted[idx]

        else:  # use smallest distance
            if abs(value - sorted_array[idx-1]) < abs(value - sorted_array[idx]):
                idx_nearest = idx_sorted[idx-1]
            else:
                idx_nearest = idx_sorted[idx]
    return idx_nearest


def find_contract_from_strike_or_delta(chain: SlicesChain,
                                       slice_id: str,
                                       option_type: str,
                                       strike_selection: StrikeSelection,
                                       nearest_strike_on_grid: NearestStrikeOnGrid = None,
                                       given_delta: float = None
                                       ) -> str:
    """
    find contract in the chain for a given slice for specified strike selection
    """
    if strike_selection == StrikeSelection.ATM:
        if option_type == 'P':
            contract = chain.get_atm_put_id(slice_id=slice_id, nearest_strike_on_grid=nearest_strike_on_grid)
        else:
            contract = chain.get_atm_call_id(slice_id=slice_id, nearest_strike_on_grid=nearest_strike_on_grid)
    elif strike_selection == StrikeSelection.DELTA:
        if option_type == 'P':
            contract = chain.get_put_delta_option_id(slice_id=slice_id, delta=given_delta,
                                                     nearest_strike_on_grid=nearest_strike_on_grid)
        else:
            contract = chain.get_call_delta_option_id(slice_id=slice_id, delta=given_delta,
                                                      nearest_strike_on_grid=nearest_strike_on_grid)
    else:
        raise ValueError(f"{option_type}")
    return contract


def get_flat_vol_expiry_slice(forward: float = 1750.0,
                              expiry_id: str = '1m',
                              expiry_time: pd.Timestamp = pd.Timestamp('2023-01-31 08:00:00+00:00'),
                              value_time: pd.Timestamp = pd.Timestamp('2023-01-01 08:00:00+00:00'),
                              strikes: np.ndarray = np.array([1500.0, 1750.0, 2000.0]),
                              flat_vol: float = 0.7,
                              discfactor: float = 1.0
                              ) -> ExpirySlice:

    ttm = compute_time_to_maturity(maturity_time=expiry_time, value_time=value_time)
    undelying_data = pd.Series({UnderlyingColumn.EXPIRY_ID: expiry_id,
                                UnderlyingColumn.VALUE_TIME: value_time,
                                UnderlyingColumn.EXPIRY: expiry_time,
                                UnderlyingColumn.FORWARD_PRICE: forward,
                                UnderlyingColumn.TTM: ttm})

    index = [f"P-{x:0.0f}" for x in strikes] + [f"C-{x:0.0f}" for x in strikes]
    optiontypes = np.concatenate((np.full(strikes.shape, 'P'), np.full(strikes.shape, 'C')))
    strikes = np.concatenate((strikes, strikes))
    vols = flat_vol*np.ones_like(strikes)

    options = pd.DataFrame(index=index)
    options[SliceColumn.CONTRACT] = index
    options[SliceColumn.STRIKE] = strikes
    options[SliceColumn.OPTION_TYPE] = optiontypes
    options[SliceColumn.MARK_IV] = vols
    options[SliceColumn.MARK_PRICE] = bsm.compute_bsm_vanilla_slice_prices(ttm=ttm,
                                                                           forward=forward,
                                                                           strikes=strikes,
                                                                           vols=vols,
                                                                           optiontypes=optiontypes,
                                                                           discfactor=discfactor)
    options[SliceColumn.USD_MULTIPLIER] = 1.0
    return ExpirySlice(undelying_data=undelying_data, options_df=options)


def get_contract_execution_price(contract_data: pd.DataFrame,
                                 num_contracts: pd.Series,
                                 is_trade_at_bid_ask: bool = False,
                                 traded_premium_charge: float = None
                                 ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series]:
    """
    compute the execution price from contract data
    """
    # align
    contract_data = contract_data.loc[num_contracts.index, :]

    mark_price = contract_data[SliceColumn.MARK_PRICE].to_numpy()
    bid_price = contract_data[SliceColumn.BID_PRICE].to_numpy()
    ask_price = contract_data[SliceColumn.ASK_PRICE].to_numpy()

    # mid_price = contract_data[SliceColumn.MARK_PRICE]
    # we take mid if both bid and ask exists, otherwise we take mid
    mid_price = pd.Series(np.where(np.logical_and(np.isnan(bid_price==False), np.isnan(ask_price==False)),
                                   0.5*(bid_price+ask_price),
                                   mark_price),
                          index=contract_data.index)

    if is_trade_at_bid_ask:
        execution_price_coin = pd.Series(np.where(np.greater(num_contracts, 0.0),
                                                  ask_price,
                                                  bid_price),
                                         index=contract_data.index)
    else:
        execution_price_coin = mid_price
    if traded_premium_charge is not None:
        # for buys increase, for sell decrease
        execution_price_coin = pd.Series(np.where(np.greater(num_contracts, 0.0),
                                                  execution_price_coin*(1.0+traded_premium_charge),  #  buy at higher price
                                                  execution_price_coin*(1.0-traded_premium_charge)),  # sell at lower price
                                         index=contract_data.index)

    slippage_coin = np.abs(execution_price_coin - mid_price)

    execution_price_usd = execution_price_coin * contract_data[SliceColumn.USD_MULTIPLIER]
    slippage_usd = slippage_coin * contract_data[SliceColumn.USD_MULTIPLIER]

    return execution_price_coin, execution_price_usd, slippage_coin, slippage_usd


class UnitTests(Enum):
    EXPIRY_SLICE_DATA = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.EXPIRY_SLICE_DATA:
        expiry_slice = get_flat_vol_expiry_slice()
        expiry_slice.print()

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.EXPIRY_SLICE_DATA

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
fetch deribit data
"""
import qis
import requests
import pandas as pd
import numpy as np
from typing import List, Literal, Union, Dict, Any
from enum import Enum
from tqdm import tqdm

# internal
from option_chain_analytics.option_chain import ExpirySlice, SliceColumn, UnderlyingColumn
from option_chain_analytics.data.config import TIME_FMT, EXPIRY_DATE_FORMAT, compute_time_to_maturity

# local paths
import local_path as lp


DERIBIT_LOCAL_PATH = f"{lp.get_resource_path()}\\deribit\\"


def get_deribit_local_file_path(current_time: pd.Timestamp,
                                ticker: Literal['BTC', 'ETH'] = 'ETH',
                                local_path: str = DERIBIT_LOCAL_PATH
                                ) -> str:
    file_path = f"{local_path}{ticker}_{current_time.strftime(TIME_FMT)}.csv"
    return file_path


def get_deribit_appended_file_path(ticker: Literal['BTC', 'ETH'] = 'ETH',
                                   local_path: str = DERIBIT_LOCAL_PATH
                                   ) -> str:
    file_path = f"{local_path}{ticker}_appended_options.feather"
    return file_path


class DeribitApi:
    """
    fetch all instruments data from deribit
    """
    def __init__(self, currency: Literal['BTC', 'ETH'] = 'BTC'):
        """
        implement as Data fetching from Deribit using api
        """
        self.url = 'https://www.deribit.com/api/v2/public/'
        self.currency = str.lower(currency)

    def get_live_instruments(self) -> pd.DataFrame:
        """
        This method is used to retrieve live instruments
        -------------
        df = api.get_live_instruments()
        df.columns = Index(['tick_size', 'taker_commission', 'strike', 'settlement_period',
       'settlement_currency', 'rfq', 'quote_currency', 'price_index',
       'option_type', 'min_trade_amount', 'maker_commission', 'kind',
       'is_active', 'instrument_name', 'instrument_id', 'expiration_timestamp',
       'creation_timestamp', 'counter_currency', 'contract_size',
       'block_trade_tick_size', 'block_trade_min_trade_amount',
       'block_trade_commission', 'base_currency', 'max_liquidation_commission',
       'max_leverage', 'future_type'],
        """
        data = {'currency': self.currency}
        df = None
        for attempt in range(10):  # tend to break on several request
            r = requests.get(f"{self.url}get_instruments", data).json()
            if 'result' in r.keys():
                df = pd.DataFrame(r['result'])
                break
            else:
                print(f"try attempt {attempt+1}: {r}")
        if df is None:
            raise ValueError(f"could not get data after 10 attempts")
        return df

    def get_instruments_urls(self) -> List[str]:
        """
         api.get_instruments_urls()
        ['https://www.deribit.com/api/v2/public/get_order_book?instrument_name=BTC-4SEP20-13250-P',
        'https://www.deribit.com/api/v2/public/get_order_book?instrument_name=BTC-26MAR21-8000-P',
        ....]
        """
        live_instruments = self.get_live_instruments()
        print(live_instruments)
        request_url = f"{self.url}get_order_book?instrument_name="
        url_storage = [f"{request_url}{instrument}" for instrument in live_instruments['instrument_name']]
        return url_storage

    def request_get(self, url) -> dict:
        """
        An intermediate function used in conjunction with the `collect_data` method.
        """
        out = None
        for attempt in range(10):  # tend to break on several request
            r = requests.get(url).json()
            if 'result' in r.keys():
                out = r['result']
                break
        if out is None:
            print(f"could not get data for {url} after 10 attempts")
        return out

    def fetch_live_data(self) -> pd.DataFrame:
        """
        Retrieves instrument data
        -------------
        df = data.collect_data()
        df.columns
        Index(['expiration_timestamp', 'option_type', 'instrument_name', 'strike',
               'underlying_price', 'underlying_index', 'timestamp', 'stats', 'state',
               'settlement_price', 'open_interest', 'min_price', 'max_price',
               'mark_price', 'mark_iv', 'last_price', 'interest_rate',
               'instrument_name', 'index_price', 'greeks', 'estimated_delivery_price',
               'change_id', 'bids', 'bid_iv', 'best_bid_price', 'best_bid_amount',
               'best_ask_price', 'best_ask_amount', 'asks', 'ask_iv'],
                dtype='object')
        """
        live_instruments = self.get_live_instruments().set_index('instrument_name')
        raw_data = []
        request_url = f"{self.url}get_order_book?instrument_name="
        for instrument in tqdm(live_instruments.index):
            url_instrument = f"{request_url}{instrument}"
            raw_data_ = self.request_get(url_instrument)
            if raw_data_ is not None:
                raw_data.append(raw_data_)
        df = pd.DataFrame(raw_data).set_index('instrument_name')

        # stats = pd.json_normalize(df['stats']).set_index(df.index)
        # print(stats)
        # df = pd.concat([df.drop('stats', axis=1), pd.json_normalize(df['stats'])], axis=1)
        # print(df)

        df_joint = pd.concat([df, live_instruments], axis=1)

        return df_joint


def parse_deribit_options_data(df: pd.DataFrame,
                               value_time: pd.Timestamp,
                               ticker: str
                               ) -> pd.DataFrame:
    """
    take deribit instruments df and produce df with SliceColumn for options data
    """
    # 1 get all options:
    option_df = df.loc[df['kind'] == 'option'].copy()
    option_df['expiry_time'] = [pd.to_datetime(x, unit='ms', utc=True) for x in option_df['expiration_timestamp']]
    option_df[SliceColumn.EXCHANGE_TIME.value] = value_time
    option_df['ttm'] = option_df.apply(lambda x: compute_time_to_maturity(x['expiry_time'], x[SliceColumn.EXCHANGE_TIME.value]), axis=1)

    # greeks string to dict and to pd.Dataframe
    greeks = pd.DataFrame.from_dict({key: x for key, x in zip(option_df.index, option_df['greeks'].to_numpy())}, orient='index')  #.apply(ast.literal_eval).to_dict()
    stats = pd.DataFrame.from_dict({key: x for key, x in zip(option_df.index, option_df['stats'].to_numpy())}, orient='index')
    # filter to include all columns
    new_options_df = pd.concat([pd.Series(option_df.index, index=option_df.index, name=SliceColumn.CONTRACT.value),
                                pd.Series(value_time, index=option_df.index, name=SliceColumn.EXCHANGE_TIME.value),
                                option_df['underlying_index'].rename(SliceColumn.UNDERLYING_INDEX.value),
                                option_df['underlying_price'].rename(SliceColumn.UNDERLYING_PRICE.value),
                                option_df['underlying_price'].rename(SliceColumn.USD_MULTIPLIER.value),
                                option_df['mark_price'].rename(SliceColumn.MARK_PRICE.value),
                                option_df['best_bid_price'].rename(SliceColumn.BID_PRICE.value),
                                option_df['best_ask_price'].rename(SliceColumn.ASK_PRICE.value),
                                option_df['best_bid_amount'].rename(SliceColumn.BID_SIZE.value),
                                option_df['best_ask_amount'].rename(SliceColumn.ASK_SIZE.value),
                                0.01 * option_df['mark_iv'].rename(SliceColumn.MARK_IV.value),
                                0.01 * option_df['bid_iv'].rename(SliceColumn.BID_IV.value),
                                0.01 * option_df['ask_iv'].rename(SliceColumn.ASK_IV.value),
                                greeks['delta'].rename(SliceColumn.DELTA.value),
                                greeks['vega'].rename(SliceColumn.VEGA.value),
                                greeks['theta'].rename(SliceColumn.THETA.value),
                                greeks['gamma'].rename(SliceColumn.GAMMA.value),
                                option_df['open_interest'].rename(SliceColumn.OPEN_INTEREST.value),
                                stats['volume'].rename(SliceColumn.VOLUME.value),
                                option_df['expiry_time'].apply(lambda x: x.strftime('%d%b%Y')).rename(SliceColumn.MATURITY_ID.value),
                                option_df['strike'].rename(SliceColumn.STRIKE.value),
                                option_df['option_type'].map({'call': 'C', 'put': 'P'}).rename(SliceColumn.OPTION_TYPE.value),
                                option_df['expiry_time'].rename(SliceColumn.EXPIRY.value),
                                option_df['ttm'].rename(SliceColumn.TTM.value),
                                pd.Series(0.0, index=option_df.index, name=SliceColumn.INTEREST_RATE.value)
                                ], axis=1)
    new_options_df[SliceColumn.CONTRACT_SIZE.value] = 0.1 if ticker == 'BTC' else 1.0
    new_options_df = new_options_df.reset_index(drop=True)
    # make sure all columns in SliceColumn exist
    new_options_df = new_options_df[[x.value for x in SliceColumn]]
    return new_options_df


def update_deribit_options_data(tickers: List[str] = ("ETH", "BTC"), is_print: bool = False) -> pd.Timestamp:
    """
    fetch live deribit data and append to database
    """
    current_time = pd.Timestamp.utcnow()  # this is ticker
    print(f"starting deribit update at {current_time}")
    for ticker in tickers:
        # timestamps[ticker] = current_time
        df = DeribitApi(ticker).fetch_live_data()
        file_path = get_deribit_local_file_path(current_time=current_time, ticker=ticker)
        # store raw df as file
        qis.save_df_to_csv(df=df, local_path=file_path)
        parsed = parse_deribit_options_data(ticker=ticker, df=df, value_time=current_time)
        # append pars to existing data
        file_path = get_deribit_appended_file_path(ticker=ticker)
        qis.append_df_to_feather(df=parsed, local_path=file_path, index_col=None)
        if is_print:
            print(f"Data saved for {ticker}")
    return current_time


def load_deribit_contract_ts_data(ticker: Union[str, Literal['BTC', 'ETH']] = 'BTC',
                                  local_path: str = DERIBIT_LOCAL_PATH
                                  ) -> Dict[str, Any]:

    file_path = get_deribit_appended_file_path(ticker=ticker, local_path=local_path)
    chain_ts = qis.load_df_from_feather(local_path=file_path, index_col=None)
    spot_data = qis.load_df_from_feather(file_name=f"{ticker}_perp_data", local_path=f"{lp.get_resource_path()}\\tardis\\")
    return dict(chain_ts=chain_ts, spot_data=spot_data, ticker=ticker)


class UnitTests(Enum):
    FILE_PATH = 1
    UPDATE_OPTIONS_DATA = 2
    LOAD_DERIBIT_OPTIONS_DF = 3


def run_unit_test(unit_test: UnitTests):

    pd.set_option('display.max_columns', 500)

    if unit_test == UnitTests.FILE_PATH:
        file_path = get_deribit_appended_file_path(ticker='BTC')
        print(file_path)

    elif unit_test == UnitTests.UPDATE_OPTIONS_DATA:
        timestamps = update_deribit_options_data()
        print(timestamps)

    elif unit_test == UnitTests.LOAD_DERIBIT_OPTIONS_DF:
        from option_chain_analytics.data.chain_ts import OptionsDataDFs
        from option_chain_analytics.data.chain_loader_from_ts import create_chain_from_from_options_dfs

        options_data_dfs = OptionsDataDFs(**load_deribit_contract_ts_data(ticker='ETH'))
        options_data_dfs.print()
        print(options_data_dfs.chain_ts.columns)
        time_index = options_data_dfs.get_timeindex()
        print(f"time_index={time_index}")

        value_time = pd.Timestamp('2023-10-27 06:20:03.160939+00:00')
        chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)
        chain.print_slices_id()


if __name__ == '__main__':

    unit_test = UnitTests.LOAD_DERIBIT_OPTIONS_DF

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import pandas as pd
import numpy as np
import yfinance as yf
import qis
from typing import List, Dict, Any, Literal, Tuple
from enum import Enum

# internal
import option_chain_analytics.pricers.bsm as bsm

# prop
from option_chain_analytics.data.config import TIME_FMT, compute_time_to_maturity
from option_chain_analytics.option_chain import SliceColumn

import local_path as local_path
YAHOO_LOCAL_PATH = f"{local_path.get_resource_path()}\\yahoo_options\\"
YAHOO_HF_LOCAL_PATH = f"{local_path.get_resource_path()}\\yahoo_hf\\"


def get_yahoo_local_file_path(current_time: pd.Timestamp,
                              ticker: str = 'SPY',
                              local_path: str = YAHOO_LOCAL_PATH
                              ) -> str:
    file_path = f"{local_path}{ticker}_{current_time.strftime(TIME_FMT)}.csv"
    return file_path


def get_yahoo_appended_file_path(ticker: str = 'SPY',
                                 local_path: str = YAHOO_LOCAL_PATH
                                 ) -> str:
    file_path = f"{local_path}{ticker}_appended_options.feather"
    return file_path


def get_yahoo_hf_appended_file_path(ticker: str = 'SPY',
                                    interval: Literal['1d', '1h', '30m', '15m', '5m', '1m'] = '30m',
                                    local_path: str = YAHOO_HF_LOCAL_PATH
                                    ) -> str:
    file_path = f"{local_path}{ticker}_{interval}.feather"
    return file_path


def fetch_yahoo_options_live_data(ticker: str = 'SPY',
                                  value_time: pd.Timestamp = pd.Timestamp.utcnow()
                                  ) -> pd.DataFrame:
    asset = yf.Ticker(ticker)
    hist = asset.history(period="1d", interval="1m")
    current_price = hist['Close'][-1]

    # for the rate use 13w bill ticker ^IRX
    rhist = yf.Ticker('^IRX').history(period="2d", interval="1m")
    discount_rate = rhist['Close'][-1] / 100.0

    def infer_forward_discount(call0: float, call1: float, put0: float, put1: float, strike0: float, strike1: float):
        discount = - ((call0 - put0) - (call1 - put1)) / (strike0 - strike1)
        forward = 0.5 * ((call0 - put0) + (call1 - put1)) / discount + 0.5 * (strike0 + strike1)
        return forward, discount

    # imply forwards
    def imply_forwards(calls: pd.DataFrame, puts: pd.DataFrame, ttm: float, last_forward: float,
                       last_discfactor: float,
                       widths: float = 0.05
                       ) -> Tuple[float, float]:
        joint_strikes = list(set(calls.index.to_list()) & set(puts.index.to_list()))
        cond = np.logical_and(np.greater(joint_strikes, (1.0-widths) * last_forward),
                              np.less(joint_strikes, (1.0+widths) * last_forward))
        if len(joint_strikes) == 0:
            return np.nan
        atm_strikes = pd.Series(joint_strikes, index=joint_strikes).where(cond)
        atm_strikes = atm_strikes.dropna()
        calls = calls.loc[atm_strikes, :]  # alighn
        puts = puts.loc[atm_strikes, :]  # alighn
        strikes = atm_strikes.to_numpy()
        ask_call, bid_call = calls['ask'].to_numpy(), calls['bid'].to_numpy()
        ask_put, bid_put = puts['ask'].to_numpy(), puts['bid'].to_numpy()

        mid_call = 0.5*(ask_call + bid_call)
        mid_put = 0.5*(ask_put + bid_put)

         # find the spread
        spread = mid_put - mid_call
        idx = np.where(np.diff(np.sign(spread)) != 0)[0] + 1
        idx = idx[0]
        forward, discount = infer_forward_discount(call0=mid_call[idx - 1], call1=mid_call[idx],
                                                   put0=mid_put[idx - 1], put1=mid_put[idx],
                                                   strike0=strikes[idx - 1], strike1=strikes[idx])

        return forward, discount

    all_options = []
    last_forward = current_price
    last_discfactor = None
    for expiry in asset.options:
        opt = asset.option_chain(expiry)
        calls, puts = opt.calls.set_index('strike', drop=False), opt.puts.set_index('strike', drop=False)
        if calls.empty or puts.empty:
            continue
        calls.index.name, puts.index.name = '', ''

        # ttm
        expiry_time = pd.Timestamp(expiry, tz='UTC').replace(hour=20)  # expire at 20.00 UTC time = 16.00 US local time
        ttm = compute_time_to_maturity(maturity_time=expiry_time, value_time=value_time)
        last_discfactor = last_discfactor or np.exp(-discount_rate * ttm)

        forward, discfactor = imply_forwards(calls=calls, puts=puts, last_forward=last_forward, last_discfactor=last_discfactor, ttm=ttm)

        discount_rate = - np.log(discfactor) / ttm
        last_forward = forward
        last_discfactor = discfactor

        #if np.isnan(forward):  # try wider width
        #    raise ValueError(f"no forward")

        # option vol
        calls, puts = opt.calls.set_index('contractSymbol'), opt.puts.set_index('contractSymbol')

        # bid-ask can be zero if outside hours
        calls_bid = calls['bid'].fillna(0.0).to_numpy()
        calls_ask = calls['ask'].fillna(0.0).to_numpy()
        calls_mark = np.where(np.logical_and(calls_bid > 0.0, calls_ask > 0.0), 0.5*(calls_bid+calls_ask),
                              np.where(calls_bid > 0.0, calls_bid, calls_ask))
        calls[SliceColumn.MARK_PRICE.value] = calls_mark

        puts_bid = puts['bid'].fillna(0.0).to_numpy()
        puts_ask = puts['ask'].fillna(0.0).to_numpy()
        puts_mark = np.where(np.logical_and(puts_bid > 0.0, puts_ask > 0.0), 0.5*(puts_bid+puts_ask),
                             np.where(puts_bid > 0.0, puts_bid, puts_ask))
        puts[SliceColumn.MARK_PRICE.value] = puts_mark

        calls[SliceColumn.BID_IV.value] = bsm.infer_bsm_ivols_from_model_slice_prices(ttm=ttm, forward=forward,
                                                                                      strikes=calls['strike'].to_numpy(),
                                                                                      optiontypes=np.full(calls.index.shape, 'C'),
                                                                                      model_prices=calls_bid,
                                                                                      discfactor=discfactor)
        calls[SliceColumn.ASK_IV.value] = bsm.infer_bsm_ivols_from_model_slice_prices(ttm=ttm, forward=forward,
                                                                                      strikes=calls['strike'].to_numpy(),
                                                                                      optiontypes=np.full(calls.index.shape, 'C'),
                                                                                      model_prices=calls_ask,
                                                                                      discfactor=discfactor)
        calls[SliceColumn.MARK_IV.value] = bsm.infer_bsm_ivols_from_model_slice_prices(ttm=ttm, forward=forward,
                                                                                       strikes=calls['strike'].to_numpy(),
                                                                                       optiontypes=np.full(calls.index.shape, 'C'),
                                                                                       model_prices=calls_mark,
                                                                                       discfactor=discfactor)
        puts[SliceColumn.BID_IV.value] = bsm.infer_bsm_ivols_from_model_slice_prices(ttm=ttm, forward=forward,
                                                                                     strikes=puts['strike'].to_numpy(),
                                                                                     optiontypes=np.full(puts.index.shape, 'P'),
                                                                                     model_prices=puts_bid,
                                                                                     discfactor=discfactor)
        puts[SliceColumn.ASK_IV.value] = bsm.infer_bsm_ivols_from_model_slice_prices(ttm=ttm, forward=forward,
                                                                                     strikes=puts['strike'].to_numpy(),
                                                                                     optiontypes=np.full(puts.index.shape, 'P'),
                                                                                     model_prices=puts_ask,
                                                                                     discfactor=discfactor)
        puts[SliceColumn.MARK_IV.value] = bsm.infer_bsm_ivols_from_model_slice_prices(ttm=ttm, forward=forward,
                                                                                      strikes=puts['strike'].to_numpy(),
                                                                                      optiontypes=np.full(puts.index.shape, 'P'),
                                                                                      model_prices=puts_mark,
                                                                                      discfactor=discfactor)

        calls[SliceColumn.OPTION_TYPE.value] = 'C'
        puts[SliceColumn.OPTION_TYPE.value] = 'P'
        option_df = pd.concat([calls, puts], axis=0)

        # enter extra data
        option_df[SliceColumn.UNDERLYING_PRICE.value] = forward
        option_df[SliceColumn.CONTRACT.value] = option_df.index.to_list()

        option_df = option_df.rename({'bid': SliceColumn.BID_PRICE.value,
                                      'ask': SliceColumn.ASK_PRICE.value,
                                      'openInterest': SliceColumn.OPEN_INTEREST.value,
                                      'volume': SliceColumn.VOLUME.value,
                                      'strike': SliceColumn.STRIKE.value}, axis=1)

        option_df[SliceColumn.EXPIRY.value] = expiry_time
        option_df[SliceColumn.TTM.value] = ttm

        if isinstance(discount_rate, np.ndarray):
            discount_rate = discount_rate[0]
        option_df[SliceColumn.INTEREST_RATE.value] = discount_rate

        option_df = option_df.drop(['lastTradeDate', 'lastPrice', 'change', 'percentChange', 'impliedVolatility',
                                    'inTheMoney', 'contractSize', 'currency'], axis=1)

        # add greeks
        strike = option_df[SliceColumn.STRIKE.value].to_numpy()
        vol = option_df[SliceColumn.MARK_IV.value].to_numpy()
        optiontype = option_df[SliceColumn.OPTION_TYPE.value].to_numpy()
        option_df[SliceColumn.DELTA.value] = bsm.compute_bsm_vanilla_delta_vector(forward=forward, ttm=ttm, strike=strike, vol=vol,
                                                     optiontype=optiontype, discfactor=discfactor)
        option_df[SliceColumn.VEGA.value] = bsm.compute_bsm_vanilla_vega_vector(forward=forward, ttm=ttm, strike=strike, vol=vol)
        option_df[SliceColumn.THETA.value] = bsm.compute_bsm_vanilla_theta_vector(forward=forward, ttm=ttm, strike=strike, vol=vol,
                                                     optiontype=optiontype, discfactor=discfactor, discount_rate=discount_rate)
        option_df[SliceColumn.GAMMA.value] = bsm.compute_bsm_vanilla_gamma_vector(forward=forward, ttm=ttm, strike=strike, vol=vol)

        all_options.append(option_df)
    df = pd.concat(all_options, axis=0)

    df[SliceColumn.MATURITY_ID.value] = df[SliceColumn.EXPIRY.value].apply(lambda x: x.strftime('%d%b%Y'))
    df[SliceColumn.USD_MULTIPLIER.value] = 1.0
    df[SliceColumn.CONTRACT.value] = df.index
    df[SliceColumn.EXCHANGE_TIME.value] = value_time
    df[SliceColumn.UNDERLYING_INDEX.value] = ticker
    df[SliceColumn.CONTRACT_SIZE.value] = 100.0
    df[SliceColumn.ASK_SIZE.value] = 1.0
    df[SliceColumn.BID_SIZE.value] = 1.0

    df = df.reset_index(drop=True)

    # make sure all columns in SliceColumn exist
    df = df[[x.value for x in SliceColumn]]

    return df


def update_options_data(tickers: List[str] = ("SPY", ), is_live_markets: bool = True) -> None:

    for ticker in tickers:
        if is_live_markets:
            value_time = pd.Timestamp.utcnow() - pd.Timedelta(minutes=20)
        else:  # yesterday close
            value_time = (pd.Timestamp.utcnow() - pd.Timedelta(days=1)).normalize().replace(hour=20)
        df = fetch_yahoo_options_live_data(ticker, value_time=value_time)
        # file_path = get_yahoo_local_file_path(current_time=value_time, ticker=ticker)
        # qis.save_df_to_csv(df=df, local_path=file_path)
        file_path = get_yahoo_appended_file_path(ticker=ticker)
        qis.append_df_to_feather(df=df, local_path=file_path, index_col=None, keep='last')

        print(f"Data saved for {ticker} at value_time={value_time}")


def fetch_hf_ohlc(ticker: str = 'SPY',
                  interval: Literal['1d', '1h', '30m', '15m', '5m', '1m'] = '30m'
                  ) -> pd.DataFrame:
    """
    fetch hf data using yf
    for m and h frequencies we shift the data forward because yf
    reports timestamps of bars at the start of the period: we shift it to the end of the period
    """
    asset = yf.Ticker(ticker)
    if interval == '1d':  # close to close
        # ohlc_data = asset.history(period="730d", interval='1d')
        ohlc_data = yf.download(tickers=ticker, start=None, end=None, ignore_tz=True)
        ohlc_data.index = ohlc_data.index.tz_localize('UTC')
    elif interval == '1h':
        ohlc_data = asset.history(period="730d", interval="1h")
        ohlc_data.index = [t + pd.Timedelta(minutes=60) for t in ohlc_data.index]
    elif interval == '30m':
        ohlc_data = asset.history(period="60d", interval="30m")
        ohlc_data.index = [t + pd.Timedelta(minutes=30) for t in ohlc_data.index]
    elif interval == '15m':
        ohlc_data = asset.history(period="60d", interval="15m")
        ohlc_data.index = [t + pd.Timedelta(minutes=15) for t in ohlc_data.index]
    elif interval == '5m':
        ohlc_data = asset.history(period="60d", interval="5m")
        ohlc_data.index = [t + pd.Timedelta(minutes=5) for t in ohlc_data.index]
    elif interval == '1m':
        ohlc_data = asset.history(period="7d", interval="1m")
        ohlc_data.index = [t + pd.Timedelta(minutes=1) for t in ohlc_data.index]
    else:
        raise NotImplementedError(f"interval={interval}")
    ohlc_data = ohlc_data.rename({'Open': 'open', 'High': 'high', 'Low': 'low', 'Close': 'close'}, axis=1)
    ohlc_data.index = ohlc_data.index.tz_convert('UTC')
    ohlc_data.index.name = 'timestamp'
    return ohlc_data


def update_hf_data(tickers: List[str] = ("SPY", ),
                   intervals: List[str] = ['1d', '1h', '30m', '15m', '5m', '1m']
                   ):
    for ticker in tickers:
        for interval in intervals:
            df = fetch_hf_ohlc(ticker=ticker, interval=interval)
            file_path = get_yahoo_hf_appended_file_path(ticker=ticker, interval=interval)
            qis.append_df_to_feather(df=df, local_path=file_path, index_col='timestamp')


def load_contract_ts_data(ticker: str = 'SPY',
                          local_path: str = YAHOO_LOCAL_PATH
                          ) -> Dict[str, Any]:

    file_path = get_yahoo_appended_file_path(ticker=ticker, local_path=local_path)
    chain_ts = qis.load_df_from_feather(local_path=file_path, index_col=None)
    # spot_data = qis.load_df_from_feather(file_name=f"{ticker}_perp_data", local_path=f"{lp.get_resource_path()}\\tardis\\")
    return dict(chain_ts=chain_ts, spot_data=None, ticker=ticker)


class UnitTests(Enum):
    UPDATE_OPTIONS_DATA = 1
    UPDATE_HF_DATA = 2
    LOAD_OPTIONS_DATA = 3


def run_unit_test(unit_test: UnitTests):

    etfs = ["SPY", "QQQ", "IWM", "HYG", "^VIX", "VXX", "EEM", "SQQQ", "TQQQ", "GLD", "USO", "TLT", "USO"]
    stocks = ["TSLA", "AAPL", "AMZN", "META", "AMD"]
    # tickers = ["EEM"]
    tickers = etfs + stocks

    if unit_test == UnitTests.UPDATE_OPTIONS_DATA:
        tickers = ['SPY']
        update_options_data(tickers, is_live_markets=False)

    elif unit_test == UnitTests.UPDATE_HF_DATA:
        update_hf_data(tickers)

    elif unit_test == UnitTests.LOAD_OPTIONS_DATA:
        from option_chain_analytics.data.chain_ts import OptionsDataDFs
        from option_chain_analytics.data.chain_loader_from_ts import create_chain_from_from_options_dfs
        options_data_dfs = OptionsDataDFs(**load_contract_ts_data(ticker='SPY'))
        options_data_dfs.print()
        time_index = options_data_dfs.get_timeindex()
        print(f"time_index={time_index}")


if __name__ == '__main__':

    unit_test = UnitTests.UPDATE_OPTIONS_DATA

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
analytics for generation of roll maturities for options rolls into front-coming Fridays
"""
# packages
from __future__ import annotations
import pandas as pd
from typing import Dict, List, Union
import qis
from qis import TimePeriod
from enum import Enum

from option_chain_analytics.data.chain_loader_from_ts import create_chain_from_from_options_dfs
from option_chain_analytics.data.chain_ts import OptionsDataDFs


class RollMaturitySelection(Enum):
    WEEKLY_FRIDAY = 1   # Friday roll to next Friday expiry
    BIWEEKLY_FRIDAY = 2   # By-weekly Friday roll to 2w next Friday expiry
    MONTHLY_LAST_FRIDAY = 3  # Monthly Friday roll to next month Friday expiry
    BUSINESS_DAYS_1W_FRIDAY = 4  # Every business day roll referencing next week Friday expiry
    BUSINESS_DAYS_2W_FRIDAY = 5  # Every business day roll referencing upto next two Friday expiries
    BUSINESS_DAYS_1M_FRIDAY = 6  # Every business day roll referencing upto next one month Friday expiries
    WEEKLY_FRIDAY_1M_FRIDAY = 7  # Every friday roll referencing next weekly Fridays expiries upto next one month Friday
    WEEKLY_EX_1W_FRIDAY_1M_FRIDAY = 8  # Every friday roll referencing next second weekly Fridays expiries upto next one month Friday
    QUARTERLY_LAST_FRIDAY = 9  # every quarter
    CALENDAR_DAY_OVERNIGHT_ROLL = 10  # overnight rolls


def get_next_roll_maturities(value_time: pd.Timestamp,
                             maturity_selection: RollMaturitySelection = RollMaturitySelection.WEEKLY_FRIDAY,
                             hour_offset: int = 8,  # corresponds to the hour of roll implementation
                             min_days_to_next_friday: int = 7,
                             min_days_to_next_quarter: int = 30,
                             max_days_for_monthly_roll: int = 42
                             ) -> Union[List[pd.Timestamp], pd.DatetimeIndex]:
    """
    generate a schedule of fridays for next rolls
    """
    if maturity_selection == RollMaturitySelection.WEEKLY_FRIDAY:
        time_period = TimePeriod(value_time + pd.offsets.Day(min_days_to_next_friday), value_time + pd.offsets.Day(21))
        next_fridays = qis.generate_dates_schedule(time_period=time_period, freq='W-FRI', hour_offset=hour_offset)
        mat_dates = [next_fridays[0]]

    elif maturity_selection == RollMaturitySelection.BIWEEKLY_FRIDAY:
        time_period = TimePeriod(value_time + pd.offsets.Day(2 * min_days_to_next_friday), value_time + pd.offsets.Day(21))
        next_fridays = qis.generate_dates_schedule(time_period=time_period, freq='2W-FRI', hour_offset=hour_offset)
        mat_dates = [next_fridays[0]]

    elif maturity_selection == RollMaturitySelection.MONTHLY_LAST_FRIDAY:
        time_period = TimePeriod(value_time + pd.offsets.Day(min_days_to_next_friday), value_time + pd.offsets.Day(2 * max_days_for_monthly_roll))
        next_fridays = qis.generate_dates_schedule(time_period=time_period, freq='M-FRI', hour_offset=hour_offset)
        mat_dates = [next_fridays[0]]

    elif maturity_selection == RollMaturitySelection.BUSINESS_DAYS_1W_FRIDAY:
        # this week rolls every b day to next friday epiry
        time_period = TimePeriod(value_time + pd.offsets.Day(min_days_to_next_friday), value_time + pd.offsets.Day(14))
        next_fridays = qis.generate_dates_schedule(time_period=time_period, freq='W-FRI', hour_offset=hour_offset)
        mat_dates = [next_fridays[0]]

    elif maturity_selection == RollMaturitySelection.BUSINESS_DAYS_2W_FRIDAY:
        # this week rolls every b day to next friday epiry
        time_period = TimePeriod(value_time + pd.offsets.Day(min_days_to_next_friday), value_time + pd.offsets.Day(21))
        next_fridays = qis.generate_dates_schedule(time_period=time_period, freq='W-FRI', hour_offset=hour_offset)
        mat_dates = next_fridays

    elif maturity_selection == RollMaturitySelection.BUSINESS_DAYS_1M_FRIDAY:
        time_period = TimePeriod(value_time + pd.offsets.Day(min_days_to_next_friday),
                                 value_time + pd.offsets.Day(max_days_for_monthly_roll))
        next_fridays = qis.generate_dates_schedule(time_period=time_period, freq='W-FRI', hour_offset=hour_offset)
        mat_dates = next_fridays

    elif maturity_selection == RollMaturitySelection.WEEKLY_FRIDAY_1M_FRIDAY:
        time_period = TimePeriod(value_time + pd.offsets.Day(min_days_to_next_friday),
                                 value_time + pd.offsets.Day(max_days_for_monthly_roll))
        next_fridays = qis.generate_dates_schedule(time_period=time_period, freq='W-FRI', hour_offset=hour_offset)
        mat_dates = next_fridays

    elif maturity_selection == RollMaturitySelection.WEEKLY_EX_1W_FRIDAY_1M_FRIDAY:
        time_period = TimePeriod(value_time + pd.offsets.Day(2 * min_days_to_next_friday),
                                 value_time + pd.offsets.Day(max_days_for_monthly_roll))
        next_fridays = qis.generate_dates_schedule(time_period=time_period, freq='W-FRI', hour_offset=hour_offset)
        mat_dates = next_fridays

    elif maturity_selection == RollMaturitySelection.QUARTERLY_LAST_FRIDAY:
        # min shift of 30 day to next quarter
        time_period = TimePeriod(value_time + pd.offsets.Day(min_days_to_next_quarter),
                                 value_time + pd.offsets.Day(min_days_to_next_quarter + 181))
        next_fridays = qis.generate_dates_schedule(time_period=time_period, freq='Q-FRI', hour_offset=hour_offset)
        mat_dates = [next_fridays[0]]

    elif maturity_selection == RollMaturitySelection.CALENDAR_DAY_OVERNIGHT_ROLL:
        time_period = TimePeriod(value_time + pd.offsets.Day(1), value_time + pd.offsets.Day(2))
        trading_days = qis.generate_dates_schedule(time_period=time_period, freq='D', hour_offset=hour_offset)
        mat_dates = trading_days

    else:
        raise NotImplementedError

    return mat_dates


def get_roll_maturity_slices_at_value_time(options_data_dfs: OptionsDataDFs,
                                           value_time: pd.Timestamp,
                                           maturity_selection: RollMaturitySelection = RollMaturitySelection.WEEKLY_FRIDAY,
                                           is_apply_open_interest_filter: bool = True,
                                           hour_offset: int = 8  # corresponds to the hour of roll implementation
                                           ) -> List[str]:
    """
    generate a slice ids for next rolls from value_time
    """
    mat_dates = get_next_roll_maturities(value_time=value_time,
                                         maturity_selection=maturity_selection,
                                         hour_offset=hour_offset)
    chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)
    if chain is not None:
        slice_ids = chain.get_slice_id_for_mat_dates(mat_dates, is_apply_open_interest_filter=is_apply_open_interest_filter)
    else:
        slice_ids = []

    return slice_ids


def generate_roll_maturities(options_data_dfs: OptionsDataDFs,
                             time_period: TimePeriod,
                             maturity_selection: RollMaturitySelection = RollMaturitySelection.WEEKLY_FRIDAY,
                             is_apply_open_interest_filter: bool = False,
                             hour_offset: int = 8,   # corresponds to the hour of roll implementation
                             max_num_days_for_monthly_roll: int = 42
                             ) -> Dict[pd.Timestamp, List[str]]:
    """
    generate a schedule of fridays for next week up to 1m1w and get available chains maturities
    the output is the dictionary of rebalancing timestamp and the list of maturities to roll model portolio
    does not take into account strike space only open interest
    """
    # set trading days first
    if maturity_selection == RollMaturitySelection.WEEKLY_FRIDAY:
        # add one week to include the next weekly roll
        this = time_period.shift_end_date_by_days(num_days=7, backward=False)
        trading_days = qis.generate_dates_schedule(this, freq='W-FRI', hour_offset=hour_offset)

    elif maturity_selection == RollMaturitySelection.BIWEEKLY_FRIDAY:
        # add one week to include the next weekly roll
        this = time_period.shift_end_date_by_days(num_days=14, backward=False)
        trading_days = qis.generate_dates_schedule(this, freq='2W-FRI', hour_offset=hour_offset)

    elif maturity_selection == RollMaturitySelection.MONTHLY_LAST_FRIDAY:
        this = time_period.shift_end_date_by_days(num_days=2*max_num_days_for_monthly_roll, backward=False)
        trading_days = qis.generate_dates_schedule(this, freq='M-FRI', hour_offset=hour_offset)

    elif maturity_selection == RollMaturitySelection.BUSINESS_DAYS_1W_FRIDAY:
        trading_days = qis.generate_dates_schedule(time_period, freq='B', hour_offset=hour_offset)

    elif maturity_selection == RollMaturitySelection.BUSINESS_DAYS_2W_FRIDAY:
        trading_days = qis.generate_dates_schedule(time_period, freq='B', hour_offset=hour_offset)

    elif maturity_selection == RollMaturitySelection.BUSINESS_DAYS_1M_FRIDAY:
        trading_days = qis.generate_dates_schedule(time_period, freq='B', hour_offset=hour_offset)

    elif maturity_selection == RollMaturitySelection.WEEKLY_FRIDAY_1M_FRIDAY:
        trading_days = qis.generate_dates_schedule(time_period, freq='W-FRI', hour_offset=hour_offset)

    elif maturity_selection == RollMaturitySelection.WEEKLY_EX_1W_FRIDAY_1M_FRIDAY:
        trading_days = qis.generate_dates_schedule(time_period, freq='W-FRI', hour_offset=hour_offset)

    elif maturity_selection == RollMaturitySelection.QUARTERLY_LAST_FRIDAY:
        # add one quarter to include the next weekly roll
        this = time_period.shift_end_date_by_days(num_days=91, backward=False)
        trading_days = qis.generate_dates_schedule(this, freq='Q-FRI', hour_offset=hour_offset)

    elif maturity_selection == RollMaturitySelection.CALENDAR_DAY_OVERNIGHT_ROLL:
        # generate extra days to cover the last week roll
        this = time_period.shift_end_date_by_days(num_days=2, backward=False)
        trading_days = qis.generate_dates_schedule(this, freq='D', hour_offset=hour_offset)

    else:
        raise NotImplementedError(f"maturity_selection = {maturity_selection}")

    roll_maturities = {}
    for date in trading_days:
        slice_ids = get_roll_maturity_slices_at_value_time(options_data_dfs=options_data_dfs,
                                                           value_time=date,
                                                           maturity_selection=maturity_selection,
                                                           is_apply_open_interest_filter=is_apply_open_interest_filter,
                                                           hour_offset=hour_offset)
        if len(slice_ids) > 0:
            roll_maturities[date] = slice_ids

    return roll_maturities


class UnitTests(Enum):
    ROLLS_AT_TIMESTAMP = 1
    ROLL_SLICES_AT_TIMESTAMP = 2
    ROLL_MATURITIES = 3


def run_unit_test(unit_test: UnitTests):

    ticker = 'BTC'
    from option_chain_analytics.ts_loaders import ts_data_loader_wrapper, DataSource

    options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker, data_source=DataSource.TARDIS_LOCAL))
    time_period = options_data_dfs.get_start_end_date()
    time_period = TimePeriod('2022-01-01 00:00:00+00:00', '2023-04-03 00:00:00+00:00')
    time_period.print()

    if unit_test == UnitTests.ROLLS_AT_TIMESTAMP:
        value_time = pd.Timestamp.utcnow()
        for maturity_selection in RollMaturitySelection:
            mat_dates = get_next_roll_maturities(value_time=value_time,
                                                 maturity_selection=maturity_selection,
                                                 hour_offset=11,
                                                 min_days_to_next_friday=4)
            print(f"{maturity_selection} = {mat_dates}")

    elif unit_test == UnitTests.ROLL_SLICES_AT_TIMESTAMP:
        value_time = pd.Timestamp('2023-04-26 00:00:00+00:00')
        for maturity_selection in RollMaturitySelection:
            slice_ids = get_roll_maturity_slices_at_value_time(options_data_dfs=options_data_dfs,
                                                               value_time=value_time,
                                                               maturity_selection=maturity_selection,
                                                               hour_offset=8)
            print(f"{maturity_selection} = {slice_ids}")

    elif unit_test == UnitTests.ROLL_MATURITIES:
        maturity_selection = RollMaturitySelection.QUARTERLY_LAST_FRIDAY

        roll_maturities = generate_roll_maturities(options_data_dfs=options_data_dfs,
                                                   maturity_selection=maturity_selection,
                                                   time_period=time_period)
        for k, v in roll_maturities.items():
            print(f"{k}: {v}")


if __name__ == '__main__':

    unit_test = UnitTests.ROLLS_AT_TIMESTAMP

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




from option_chain_analytics.data.__init__ import *

from option_chain_analytics.public_apis.__init__ import *

from option_chain_analytics.utils.__init__ import *

from option_chain_analytics.visuals.__init__ import *


from option_chain_analytics.option_chain import (ExpirySlice,
                                                 SliceColumn,
                                                 SlicesChain,
                                                 get_contract_execution_price)



import pandas as pd
import matplotlib.pyplot as plt
import qis
from enum import Enum

# analytics
from option_chain_analytics import (OptionsDataDFs,
                                                 create_chain_from_from_options_dfs,
                                                 plot_slice_vols,
                                                 plot_slice_open_interest,
                                                 run_chain_report)
from option_chain_analytics.ts_loaders import ts_data_loader_wrapper, DataSource


pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)


class UnitTests(Enum):
    PRINT_CHAIN_DATA = 1
    PLOT_SLICE_DATA = 2
    RUN_CHAIN_REPORT = 3


def run_unit_test(unit_test: UnitTests):

    import local_path as local_path

    ticker = 'BTC'  # BTC, ETH
    # value_time = pd.Timestamp('2023-10-01 08:00:00+00:00')
    value_time = pd.Timestamp('2023-10-01 08:00:00+00:00')

    options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker, data_source=DataSource.TARDIS_LOCAL))
    chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)

    if unit_test == UnitTests.PRINT_CHAIN_DATA:
        for expiry, eslice in chain.expiry_slices.items():
            eslice.print()

    elif unit_test == UnitTests.PLOT_SLICE_DATA:
        eslice = chain.expiry_slices['31MAR23']
        plot_slice_vols(eslice=eslice)
        plot_slice_open_interest(eslice=eslice)

    elif unit_test == UnitTests.RUN_CHAIN_REPORT:
        figs = run_chain_report(chain=chain)
        qis.save_figs_to_pdf(figs=figs,
                             file_name=f"chain_report_{value_time:%Y%m%dT%H%M%S}",
                             orientation='landscape',
                             local_path=local_path.get_output_path())

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.RUN_CHAIN_REPORT

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




import timeit
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import qis

import option_chain_analytics.pricers.bsm as bsm
from enum import Enum


def test_vector_vs_loop(size: int = 1000):
    forward = np.random.uniform(1.0, 200.0, size)
    strike = np.random.uniform(1.0, 200.0, size)
    ttm = np.random.uniform(0.0, 1.0, size)
    vol = np.random.uniform(0.1, 1.0, size)
    optiontype = np.random.choice(['C', 'P'], size)
    discfactor = np.random.uniform(0.9, 1.0, size)

    def vector_pricer():
        return bsm.compute_bsm_vanilla_price_vector(forward=forward,
                                                    strike=strike,
                                                    ttm=ttm,
                                                    vol=vol,
                                                    optiontype=optiontype,
                                                    discfactor=discfactor)

    def slice_pricer():
        return bsm.compute_bsm_vanilla_slice_prices(ttm=ttm[0],
                                                    forward=forward[0],
                                                    strikes=strike,
                                                    vols=vol,
                                                    optiontypes=optiontype,
                                                    discfactor=discfactor[0])

    def spot_grid():
        return bsm.compute_bsm_forward_grid_prices(ttm=ttm[0],
                                                   forwards=forward,
                                                   strike=strike[0],
                                                   vol=vol[0],
                                                   optiontype=optiontype[0],
                                                   discfactor=discfactor[0])

    slice_pricer()
    vector_pricer()
    spot_grid()
    n = 20
    print(f"slice_pricer: avg={timeit.Timer(slice_pricer).timeit(n)/n:.4f}")
    print(f"vector_pricer: avg={timeit.Timer(vector_pricer).timeit(n)/n:.4f}")
    print(f"spot_grid: avg={timeit.Timer(spot_grid).timeit(n)/n:.4f}")


def inverse_delta():
    ttm = 0.1
    forward = 1.0
    strikes = np.linspace(0.1, 2.0, 100)
    vols = 0.5*np.ones_like(strikes)
    optiontypes = np.full_like(strikes, 'C', dtype=str)
    prices = bsm.compute_bsm_vanilla_slice_prices(ttm=ttm,
                                                  forward=forward,
                                                  strikes=strikes,
                                                  vols=vols,
                                                  optiontypes=optiontypes)
    print(prices)
    deltas = bsm.compute_bsm_vanilla_slice_deltas(ttm=ttm,
                                                  forward=forward,
                                                  strikes=strikes,
                                                  vols=vols,
                                                  optiontypes=optiontypes)
    print(deltas)
    df = pd.DataFrame(index=strikes)
    df['delta'] = deltas
    df['pa delta'] = deltas - prices / forward
    qis.plot_line(df=df)


class UnitTests(Enum):
    VECTOR_VS_LOOP = 1
    INVERSE_DELTA = 2


def run_unit_test(unit_test: UnitTests):
    np.random.seed(3)
    if unit_test == UnitTests.VECTOR_VS_LOOP:
        test_vector_vs_loop()

    elif unit_test == UnitTests.INVERSE_DELTA:
        inverse_delta()


    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.INVERSE_DELTA

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




import pandas as pd
import qis
from typing import Tuple
from qis import timer, TimePeriod
from enum import Enum

from option_chain_analytics import OptionsDataDFs, generate_atm_vols
from option_chain_analytics.ts_loaders import ts_data_loader_wrapper, DataSource


@timer
def fetch_atm_vols(options_data_dfs: OptionsDataDFs,
                   time_period: TimePeriod = None,
                   freq: str = 'D',
                   hour_offset: int = 8,
                   days_before_roll: int = 7
                   ) -> Tuple[pd.Series, pd.Series]:
    """
    fetch atm vols and aligned prices:
    """
    atm_vols = generate_atm_vols(options_data_dfs=options_data_dfs,
                                 time_period=time_period, freq=freq, hour_offset=hour_offset,
                                 days_before_roll=days_before_roll)
    price_data = options_data_dfs.get_spot_price(index=atm_vols.index)
    return atm_vols, price_data


class UnitTests(Enum):
    FETCH_ATM_VOLS = 1


def run_unit_test(unit_test: UnitTests):

    import local_path as local_path

    ticker = 'ETH'

    if unit_test == UnitTests.FETCH_ATM_VOLS:
        time_period = TimePeriod('2023-01-01 00:00:00+00:00', '2023-01-30 00:00:00+00:00', tz='UTC')
        options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker, data_source=DataSource.TARDIS_LOCAL))
        atm_vols, price_data = fetch_atm_vols(options_data_dfs=options_data_dfs,
                                              days_before_roll=7,
                                              time_period=None)
        df = pd.concat([price_data.rename(ticker), atm_vols.rename('atm_vol')], axis=1)
        print(df)
        # qis.save_df_to_csv(df=df, file_name=f"{ticker}_atm_vols", local_path=local_path.get_resource_path())


if __name__ == '__main__':

    unit_test = UnitTests.FETCH_ATM_VOLS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




# packages
import numpy as np
import pandas as pd
from typing import Tuple

import option_chain_analytics.pricers.bsm as bsm
from option_chain_analytics.option_chain import SliceColumn


def compute_portfolio_payoff(portfolio_report: pd.DataFrame,
                             current_spot: float,
                             spot_grid: np.ndarray = None,
                             spot_position_units: float = None,
                             change: float = 0.5,
                             total_name: str = 'Total P&L',
                             is_usd_pnl: bool = True
                             ) -> Tuple[pd.DataFrame, float]:
    """
    compute option portfolio payoff on a grid
    portfolio_report has columns of data SliceColumn + sizes
    """
    if spot_grid is None:
        spot_grid = np.linspace(100.0*np.floor(current_spot*(1.0-change)/100.0), 100.0*np.ceil(current_spot*(1.0+change)/100.0), 200)
    payoff, payoff0 = np.zeros_like(spot_grid), 0.0
    for strike, optiontype, size in zip(portfolio_report[SliceColumn.STRIKE],
                                        portfolio_report[SliceColumn.OPTION_TYPE],
                                        portfolio_report['sizes']):
        if optiontype == 'C':
            payoff += size*np.maximum(spot_grid-strike, 0.0)
            payoff0 += size*np.maximum(current_spot-strike, 0.0)
        elif optiontype == 'P':
            payoff += size*np.maximum(strike-spot_grid, 0.0)
            payoff0 += size*np.maximum(strike-current_spot, 0.0)
        else:
            raise NotImplementedError
    pnl = pd.Series(payoff, index=spot_grid, name='Options Payoff')
    if spot_position_units is not None and np.abs(spot_position_units) > 0.0:
        pnl = pnl.to_frame()
        spot_pnl = (spot_grid-current_spot) * spot_position_units
        pnl['Spot P&L'] = spot_pnl
        pnl[total_name] = payoff + spot_pnl
    else:
        pnl = pnl.rename(total_name).to_frame()

    if not is_usd_pnl:
        # pnl = pnl.divide(qis.np_array_to_df_columns(spot_grid, n_col=len(pnl.columns)))
        pnl = pnl.divide(spot_grid, axis=0)
        payoff0 = payoff0 / current_spot

    return pnl, payoff0


def compute_option_portfolio_dt(portfolio_report: pd.DataFrame,
                                current_spot: float,
                                ttm: float = 0.1,
                                dt: float = 0.00,
                                spot_grid: np.ndarray = None,
                                spot_position: pd.Series = None,
                                change: float = 0.5,
                                total_name: str = 'Total P&L'
                                ) -> Tuple[pd.DataFrame, float]:
    """
    compute option portfolio value change on a spot grid and dt
    portfolio_report has columns of data SliceColumn + sizes
    """
    if spot_grid is None:
        spot_grid = np.linspace(100.0*np.floor(current_spot*(1.0-change)/100.0), 100.0*np.ceil(current_spot*(1.0+change)/100.0), 200)
    payoff, payoff0 = np.zeros_like(spot_grid), 0.0
    for strike, optiontype, size, vol, mark_price in zip(portfolio_report[SliceColumn.STRIKE],
                                                         portfolio_report[SliceColumn.OPTION_TYPE],
                                                         portfolio_report['sizes'],
                                                         portfolio_report[SliceColumn.MARK_IV],
                                                         portfolio_report[SliceColumn.MARK_PRICE]):
        grid_price = bsm.compute_bsm_forward_grid_prices(ttm=ttm - dt,
                                                         forwards=spot_grid,
                                                         strike=strike,
                                                         vol=vol,
                                                         optiontype=optiontype)
        payoff += size*(grid_price-mark_price)
    pnl = pd.Series(payoff, index=spot_grid, name='Options Payoff')
    if spot_position is not None and np.abs(spot_position.iloc[-1]) > 0.0:
        pnl = pnl.to_frame()
        spot_pnl = (spot_grid/current_spot-1.0) * spot_position.iloc[-1]
        pnl['Spot P&L'] = spot_pnl
        pnl[total_name] = payoff + spot_pnl
    else:
        pnl = pnl.rename(total_name).to_frame()

    return pnl, payoff0



import time
import pandas as pd
import numpy as np
from typing import List, Optional, Union
from enum import Enum
from cryptocmd import CmcScraper


def fetch_instrument_data(ticker: str = 'btc',
                          time_str: str = 'timestamp'
                          ) -> pd.DataFrame:
    """
    fetch using cmc scarpper
    """
    if ticker is None:
        raise TypeError(f"ticker is none")

    scraper = CmcScraper(coin_code=ticker, all_time=True, order_ascending=True)

    try:
        data = scraper.get_dataframe()
    except TypeError:
        print(f"no data for ticker = {ticker}")
        return pd.DataFrame()

    data['Date'] = pd.to_datetime(data['Date'])
    data = data.set_index('Date', drop=True)
    data.index = data.index.normalize()  #remove seconds info
    data.index.name = time_str

    column_map = {'Close': 'close',
                  'Market Cap': 'market_cap',
                  'Volume': 'volume',
                  'Open': 'open',
                  'High': 'high',
                  'Low': 'low'}
    data = data.rename(column_map, axis=1)
    data = data.replace({0.0: np.nan})

    return data


class UnitTests(Enum):
    DATA = 1


def run_unit_test(unit_test: UnitTests):

    import matplotlib.pyplot as plt

    if unit_test == UnitTests.DATA:
        ticker = 'USDC'
        data = fetch_instrument_data(ticker=ticker)
        print(data)
        data['market_cap'].plot()
        data['volume'].plot()

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.DATA

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import ccxt
import re
import numpy as np
import pandas as pd
from typing import Dict, List, Union, Tuple, Optional, Any
from enum import Enum


def get_supported_exchanges() -> List[str]:
    return ccxt.exchanges


def get_exchange(exchange: str = 'deribit') -> ccxt.Exchange:
    try:
        exchange = getattr(ccxt, exchange)()
    except AttributeError:
        raise KeyError(f"Exchange {exchange} not found")
    return exchange


def get_mid_price(exchange: str = 'binance', symbol: str = 'ETHUSDT') -> float:
    exchange = get_exchange(exchange=exchange)
    data = exchange.fetch_ticker(symbol=symbol)
    return 0.5*(data['bid']+data['ask'])


def get_available_instruments(exchange: str = 'deribit') -> Tuple[List[str], List[str]]:
    exchange = get_exchange(exchange=exchange)
    exchange.load_markets()
    return exchange.markets, exchange.markets_by_id


def fetch_symbol(exchange: str = 'okx', symbol: str = 'BTC-USDT-230331') -> Dict[str, Any]:
    exchange = get_exchange(exchange=exchange)
    return exchange.fetchTicker(symbol=symbol)


def fetch_ohlcv(symbols: List[str],
                exchange: str = 'deribit',
                freq: str = '1h',
                is_normalize_volume: bool = True
                ) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
    """
    returns open, high, low, close, volume
    Available timeframes are: 1m, 3m, 5m, 10m, 15m, 30m, 1h, 2h, 3h, 6h, 12h, 1d
    symbol = 'ETH-28JAN22' for deribit
    """
    if freq == 'H':
        freq = '1h'
    elif freq == 'D':
        freq = '1d'

    if exchange == 'huobi':
        limit = 2000
    elif exchange == 'bybit':
        limit = 200
    else:
        limit = 5000

    exchange = get_exchange(exchange=exchange)

    # Check if fetching of OHLC Data is supported
    if not exchange.has["fetchOHLCV"]:
        raise ValueError(f"{exchange} does not support fetching OHLC data")

    # Check requested timeframe is available
    if (not hasattr(exchange, 'timeframes')) or (freq not in exchange.timeframes):
        print(f"The requested timeframe {freq} is not available from {exchange}")
        print('Available timeframes are:')
        for key in exchange.timeframes.keys():
            print(key)
        raise ValueError

    dfs = {}
    columns = ['timestamp', 'open', 'high', 'low', 'close', 'volume']

    def process_data(data: pd.DataFrame) -> pd.DataFrame:
        df = pd.DataFrame(data, columns=columns)
        df[columns] = df[columns].apply(pd.to_numeric, errors='ignore')
        df['timestamp'] = pd.to_datetime(df['timestamp'], unit='ms', utc=True)
        df = df.set_index('timestamp')
        return df

    for symbol in symbols:
        try:
            data = exchange.fetch_ohlcv(symbol=symbol, timeframe=freq, limit=limit) # params=dict(include_old=True)

        except (TypeError, ccxt.errors.BadSymbol):
            print(f"{symbol} not available on {exchange}, creating empty frame")
            df = pd.DataFrame(columns=columns).set_index('timestamp')

        else:
            df = process_data(data)
            if normalize_volume:
                df = normalize_volume(exchange=exchange, symbol=symbol, df=df)

        dfs[symbol] = df

    if len(symbols) == 1:
        dfs = dfs[symbols[0]]

    return dfs


def fetch_close_prices(symbols: List[str],
                       exchange: str = 'deribit',
                       freq: str = '1h'
                       ) -> pd.DataFrame:
    """
    returns open, high, low, close, volume
    Available timeframes are: 1m, 3m, 5m, 10m, 15m, 30m, 1h, 2h, 3h, 6h, 12h, 1d
    symbol = 'ETH-28JAN22' for deribit
    """
    df = {}
    for symbol in symbols:
        df[symbol] = fetch_ohlcv(symbols=[symbol], exchange=exchange, freq=freq)['close'].rename(symbol)
    df = pd.DataFrame.from_dict(df, orient='columns')
    return df


def fetch_ohlcv_with_last(symbols: List[str],
                          exchange: str = 'deribit',
                          freq: str = '1h',
                          last_timestamp: pd.Timestamp = None,  # to sync last stamp use common last_timestamp
                          use_bid_ask_for_last: bool = True
                          ) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
    """
    call fetch_ohlcv and add most recent timestamp
    """
    dfs = fetch_ohlcv(symbols=symbols, exchange=exchange, freq=freq, is_normalize_volume=False) # will be called with new pools
    exchange = get_exchange(exchange=exchange)
    for symbol in symbols:
        if last_timestamp is not None:
            try:
                data = exchange.fetch_ticker(symbol=symbol)
            except (TypeError, ccxt.errors.BadSymbol):
                print(f"{symbol} not available on {exchange}, no last is added")
                continue
            if use_bid_ask_for_last:
                if 'bid' and 'ask' in data.keys():
                    bid, ask = data['bid'], data['ask']
                    if bid is not None and ask is not None:
                        mid = 0.5*(bid+ask)
                    else:
                        mid = data['close']
                else:
                    mid = data['close']
                data_to_enter = np.array([data['open'], data['high'], data['low'], mid, data['baseVolume']])
            else:
                data_to_enter = np.array([data['open'], data['high'], data['low'], data['close'], data['baseVolume']])

            if isinstance(dfs, dict):
                dfs[symbol].loc[last_timestamp, :] = data_to_enter
            else:
                dfs.loc[last_timestamp, :] = data_to_enter

    if isinstance(dfs, dict):
        for symbol, df in dfs.items():
            dfs[symbol] = normalize_volume(exchange=exchange, symbol=symbol, df=df)
    else:
        dfs = normalize_volume(exchange=exchange, symbol=symbols[0], df=dfs)

    return dfs


def normalize_volume(exchange: ccxt.exchanges, symbol: str, df: pd.DataFrame) -> pd.DataFrame:
    """
    normalize echange volume based on contract specs
    """
    if isinstance(exchange, ccxt.bybit):  # volumes to contract
        if bool(re.search(r'\d', symbol)):  # for termed future
            df['volume'] = df['volume'] / df['close']

    if isinstance(exchange, ccxt.okx):  # volumes to contract
        if bool(re.search(r'\d', symbol)):  # for termed future
            print(symbol)
            df['volume'] = df['volume'] * 0.01
    return df


def fetch_funding_rate(symbol: str = 'ETH/USDT:USDT',
                       exchange: str = 'binanceusdm',
                       freq: Optional[str] = 'H'
                       ) -> pd.Series:
    exchange = get_exchange(exchange)
    funding = exchange.fetch_funding_rate_history(symbol)
    df = pd.json_normalize(funding)
    df['timestamp'] = pd.to_datetime(df['timestamp'], unit='ms', utc=True)
    df = df.set_index('timestamp')
    if 'info.fundingRate' in df.columns:  # this is accurate to what is paid
        frate = df['info.fundingRate'].apply(pd.to_numeric, errors='ignore')
    else:
        frate = df['fundingRate'].apply(pd.to_numeric, errors='ignore')
    if freq is not None:
        frate = frate.resample('H').last()
    return frate


class UnitTests(Enum):
    EXCHANGES = 1
    PRICE_QUOTE = 2
    INSTRUMENTS = 3
    FETCH = 4
    FETCH_WITH_LAST = 5
    FUNDING_RATE = 6
    FETCH_TICKER = 7
    TEST = 8


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.EXCHANGES:
        exchanges = get_supported_exchanges()
        for idx, exchange in enumerate(exchanges):
            print(f"{idx+1}: {exchange}")

    elif unit_test == UnitTests.PRICE_QUOTE:
        mid_price = get_mid_price(exchange='binance', symbol='ETHUSDT')
        print(mid_price)

    elif unit_test == UnitTests.INSTRUMENTS:
        markets, markets_by_id = get_available_instruments(exchange='binance')
        asset = 'BTC'
        print("########## markets############################")
        for idx, symbol in enumerate(markets):
            if asset in symbol:
                print(f"{idx+1}: {symbol}")
        print("########## markets_by_id  ############################")
        for idx, symbol in enumerate(markets_by_id):
            if asset in symbol:
                print(f"{idx+1}: {symbol}")

    elif unit_test == UnitTests.FETCH:
        # dfs = fetch_ohlcv(symbols=['ETH-PERPETUAL', 'ETH-31MAR23', 'ETH-30JUN23'], exchange='deribit', freq='H')
        # dfs = fetch_ohlcv(symbols=['ETHUSDH23', 'ETHUSDM23'], exchange='bybit', freq='H')
        # dfs = fetch_ohlcv(symbols=['ETH/USD:USDC', 'ETH/USD:ETH-230331', 'ETH/USD:ETH-230630'], exchange='bybit', freq='1h')
        # dfs = fetch_ohlcv(symbols=['ETH-USDT-SWAP', 'ETH-USDT-230224', 'ETH-USDT-230331'], exchange='okx', freq='1h')
        # dfs = fetch_ohlcv(symbols=['BTC-USDT-230331'], exchange='okx', freq='1h')
        # dfs = fetch_ohlcv(symbols=['ETH-USD', 'ETH230113', 'ETH230331'], exchange='huobi', freq='1h')
        # dfs = fetch_ohlcv(symbols=['BTCUSDT_230331'], exchange='binanceusdm', freq='1h')
        dfs = fetch_ohlcv(symbols=['BTCUSDT'], exchange='binance', freq='1h')
        # dfs = fetch_ohlcv(symbols=['ETHUSDT'], exchange='binanceusdm', freq='1h')

        print(dfs)
        for k, v in dfs.items():
            print(k)
            # print(v)
            print(v.rolling(24).sum())

    elif unit_test == UnitTests.FETCH_WITH_LAST:
        dfs = fetch_ohlcv_with_last(symbols=['ETH-USDT-230929'], exchange='okx', freq='1h')
        print(dfs)

    elif unit_test == UnitTests.FUNDING_RATE:
        #funding = fetch_funding_rate(ticker='ETHUSDT', exchange='binanceusdm')
        # funding = fetch_funding_rate(symbol='LOOKS/USDT', exchange='okx')
        funding = fetch_funding_rate(symbol='BTCUSDT', exchange='binanceusdm')
        #funding = fetch_funding_rate(symbol='FIS/USDT', exchange='gateio')
        print(funding)

    elif unit_test == UnitTests.FETCH_TICKER:
        data = fetch_symbol(exchange='okx', symbol='ETH-USDT-230331')
        print(data)
        data = fetch_symbol(exchange='binanceusdm', symbol='BTCUSDT_230331')
        print(data)

        exchange = get_exchange(exchange='binanceusdm')
        this = exchange.fetchOpenInterestHistory('BTCUSDT_230331')
        print(this)

        exchange = get_exchange(exchange='okx')
        this = exchange.fetchOpenInterestHistory('BTCUSDT_230331')
        print(this)

    elif unit_test == UnitTests.TEST:
        symbol = 'ETH-USDT-230929'
        exchange = 'okx'
        exchange = get_exchange(exchange=exchange)
        data = exchange.fetch_ticker(symbol=symbol)
        print(data)


if __name__ == '__main__':

    unit_test = UnitTests.FUNDING_RATE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
utility functions
"""
import numpy as np
from numba import njit
from typing import Union

from scipy.integrate import quad

ONE_OVER_SQRT_TWO_PI = 0.3989422804014327
ONE_OVER_SQRT_TWO = 0.7071067811865475


@njit(cache=False, fastmath=True)
def erfcc(x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    """
    Complementary error function
    """
    z = np.abs(x)
    t = 1. / (1. + 0.5*z)
    r = t * np.exp(-z*z-1.26551223+t*(1.00002368+t*(0.37409196+t*(0.09678418+t*(-0.18628806+t*(0.27886807+
        t*(-1.13520398+t*(1.48851587+t*(-.82215223+t*0.17087277)))))))))
    fcc = np.where(np.greater(x, 0.0), r, 2.0-r)
    return fcc


@njit(cache=False, fastmath=True)
def ncdf(x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    return 1.0 - 0.5*erfcc(ONE_OVER_SQRT_TWO*x)


@njit(cache=False, fastmath=True)
def npdf(x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    return np.exp(-0.5*np.square(x))/np.sqrt(2.0*np.pi)


@njit
def range_prob_t(t: float, vol: float, xa: float, xb: float) -> float:
    vol_st = vol * np.sqrt(t)
    da = xa / vol_st + 0.5*vol_st
    db = xb / vol_st + 0.5*vol_st
    return ncdf(db) - ncdf(da)


def compute_int_range_prob(t: float, vol: float, p0: float, pa: float, pb: float):
    xa = np.log(pa / p0)
    xb = np.log(pb / p0)

    def func(t_: float) -> float:
        return range_prob_t(t=t_, vol=vol, xa=xa, xb=xb)

    return quad(func, a=1e-16, b=t)[0] / t
